import os
from pathlib import Path
from unittest.mock import patch

from ploigos_step_runner import StepResult, StepRunnerException, WorkflowResult
from ploigos_step_runner.step_implementers.package import MavenPackage
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase


@patch("ploigos_step_runner.step_implementers.shared.MavenGeneric.__init__")
class TestStepImplementerMavenPackage___init__(BaseStepImplementerTestCase):
    def test_defaults(self, mock_super_init):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        config = {}

        MavenPackage(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config
        )

        mock_super_init.assert_called_once_with(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=None,
            maven_phases_and_goals=['package']
        )

    def test_given_environment(self, mock_super_init):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        config = {}

        MavenPackage(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment='mock-env'
        )

        mock_super_init.assert_called_once_with(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment='mock-env',
            maven_phases_and_goals=['package']
        )

class TestStepImplementerMavenPackage_step_implementer_config_defaults(
    BaseStepImplementerTestCase
):
    def test_result(self):
        self.assertEqual(
            MavenPackage.step_implementer_config_defaults(),
            {
                'pom-file': 'pom.xml',
                'tls-verify': True,
                'maven-profiles': [],
                'maven-additional-arguments': [],
                'maven-no-transfer-progress': True,
                'maven-additional-arguments': [
                    '-Dmaven.test.skip=true'
                ],
                'artifact-extensions': ["jar", "war", "ear"],
                'artifact-parent-dir': 'target'
            }
        )

class TestStepImplementerMavenPackage__required_config_or_result_keys(
    BaseStepImplementerTestCase
):
    def test_result(self):
        self.assertEqual(
            MavenPackage._required_config_or_result_keys(),
            [
                'pom-file'
            ]
        )

@patch.object(MavenPackage, '_run_maven_step')
@patch.object(MavenPackage, 'write_working_file', return_value='/mock/mvn_output.txt')
class TestStepImplementerMavenPackage__run_step(
    BaseStepImplementerTestCase
):
    def create_step_implementer(
            self,
            step_config={},
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=MavenPackage,
            step_config=step_config,
            step_name='package',
            implementer='MavenPackage',
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

    def test_success_single_packaged_artifact(
        self,
        mock_write_working_file,
        mock_run_maven_step
    ):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file = os.path.join(test_dir.path, 'mock-pom.xml')
            step_config = {
                'pom-file': pom_file
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # setup sideeffects
            artifact_parent_dir = os.path.join(test_dir.path, 'target')
            package_artifact_names = [
                f'my-app.jar'
            ]
            def run_maven_side_effect(mvn_output_file_path):
                os.makedirs(artifact_parent_dir, exist_ok=True)
                for artifact_name in package_artifact_names:
                    artifact_path = os.path.join(
                        artifact_parent_dir,
                        artifact_name
                    )
                    Path(artifact_path).touch()

            mock_run_maven_step.side_effect = run_maven_side_effect

            # run step
            actual_step_result = step_implementer._run_step()

            # create expected step result
            expected_step_result = StepResult(
                step_name='package',
                sub_step_name='MavenPackage',
                sub_step_implementer_name='MavenPackage'
            )
            expected_step_result.add_artifact(
                description="Standard out and standard error from maven.",
                name='maven-output',
                value='/mock/mvn_output.txt'
            )
            expected_step_result.add_artifact(
                name='packages',
                value=[{
                    'path': os.path.join(artifact_parent_dir, 'my-app.jar')
                }]
            )

            # verify step result
            self.assertEqual(
                actual_step_result,
                expected_step_result
            )

            mock_write_working_file.assert_called_once()
            mock_run_maven_step.assert_called_with(
                mvn_output_file_path='/mock/mvn_output.txt'
            )

    def test_success_multiple_packaged_artifact(
        self,
        mock_write_working_file,
        mock_run_maven_step
    ):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file = os.path.join(test_dir.path, 'mock-pom.xml')
            step_config = {
                'pom-file': pom_file
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # setup sideeffects
            artifact_parent_dir = os.path.join(test_dir.path, 'target')
            package_artifact_names = [
                f'my-app.jar',
                f'my-app.war'
            ]
            def run_maven_side_effect(mvn_output_file_path):
                os.makedirs(artifact_parent_dir, exist_ok=True)
                for artifact_name in package_artifact_names:
                    artifact_path = os.path.join(
                        artifact_parent_dir,
                        artifact_name
                    )
                    Path(artifact_path).touch()

            mock_run_maven_step.side_effect = run_maven_side_effect

            # run step
            actual_step_result = step_implementer._run_step()

            # create expected step result
            expected_step_result = StepResult(
                step_name='package',
                sub_step_name='MavenPackage',
                sub_step_implementer_name='MavenPackage'
            )
            expected_step_result.add_artifact(
                description="Standard out and standard error from maven.",
                name='maven-output',
                value='/mock/mvn_output.txt'
            )
            expected_step_result.add_artifact(
                name='packages',
                value=[
                    {
                        'path': os.path.join(artifact_parent_dir, 'my-app.jar')
                    },
                    {
                        'path': os.path.join(artifact_parent_dir, 'my-app.war')
                    }
                ]
            )

            # verify step result
            self.assertEqual(
                actual_step_result,
                expected_step_result
            )

            mock_write_working_file.assert_called_once()
            mock_run_maven_step.assert_called_with(
                mvn_output_file_path='/mock/mvn_output.txt'
            )

    def test_fail_maven_run(
        self,
        mock_write_working_file,
        mock_run_maven_step
    ):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file = os.path.join(test_dir.path, 'mock-pom.xml')
            step_config = {
                'pom-file': pom_file
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # run step with mock failure
            mock_run_maven_step.side_effect = StepRunnerException('Mock error running maven')
            actual_step_result = step_implementer._run_step()

            # create expected step result
            surefire_reports_dir = os.path.join(test_dir.path, 'target/surefire-reports')
            expected_step_result = StepResult(
                step_name='package',
                sub_step_name='MavenPackage',
                sub_step_implementer_name='MavenPackage'
            )
            expected_step_result.success = False
            expected_step_result.message = "Error running 'maven package' to package artifacts. " \
                "More details maybe found in 'maven-output' report artifact: " \
                "Mock error running maven"
            expected_step_result.add_artifact(
                description="Standard out and standard error from maven.",
                name='maven-output',
                value='/mock/mvn_output.txt'
            )

            # verify step result
            self.assertEqual(
                actual_step_result,
                expected_step_result
            )

            mock_write_working_file.assert_called_once()
            mock_run_maven_step.assert_called_with(
                mvn_output_file_path='/mock/mvn_output.txt'
            )

    def test_fail_no_find_artifacts(
        self,
        mock_write_working_file,
        mock_run_maven_step
    ):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file = os.path.join(test_dir.path, 'mock-pom.xml')
            step_config = {
                'pom-file': pom_file,
                'artifact-parent-dir': 'mock/does-not-exist'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # run step
            actual_step_result = step_implementer._run_step()

            # create expected step result
            expected_step_result = StepResult(
                step_name='package',
                sub_step_name='MavenPackage',
                sub_step_implementer_name='MavenPackage'
            )
            expected_step_result.success = False
            expected_step_result.message = \
                "Error finding artifacts after running maven package:" \
                f" [Errno 2] No such file or directory: '{test_dir.path}/mock/does-not-exist'"
            expected_step_result.add_artifact(
                description="Standard out and standard error from maven.",
                name='maven-output',
                value='/mock/mvn_output.txt'
            )

            # verify step result
            self.assertEqual(actual_step_result,expected_step_result)

            mock_write_working_file.assert_called_once()
            mock_run_maven_step.assert_called_with(
                mvn_output_file_path='/mock/mvn_output.txt'
            )



"""`StepImplementers` for the `container-image-static-vulnerability-scan` step.
"""

from ploigos_step_runner.step_implementers.container_image_static_vulnerability_scan.\
    openscap import OpenSCAP



import os
import re
import sys
from pathlib import Path
from unittest.mock import patch

import sh
from ploigos_step_runner import StepResult
from ploigos_step_runner.step_implementers.create_container_image import \
    Buildah
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase


class BaseTestStepImplementerCreateContainerImageBuildah(BaseStepImplementerTestCase):
    def create_step_implementer(
            self,
            step_config={},
            step_name='',
            implementer='',
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=Buildah,
            step_config=step_config,
            step_name=step_name,
            implementer=implementer,
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

@patch("ploigos_step_runner.StepImplementer._validate_required_config_or_previous_step_result_artifact_keys")
class TestStepImplementerCreateContainerImageBuildah__validate_required_config_or_previous_step_result_artifact_keys(
    BaseTestStepImplementerCreateContainerImageBuildah
):
    def test_valid_defaults(self, mock_super_validate):
        with TempDirectory() as test_dir:
            # setup
            parent_work_dir_path = os.path.join(test_dir.path, 'working')
            step_config = {
                'context': test_dir.path,
                'organization': 'mock-org',
                'service-name': 'mock-service',
                'application-name': 'mock-app'
            }
            test_dir.write('Containerfile',b'''testing''')

            # run test
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )
            step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

            # validate
            mock_super_validate.assert_called_once_with()

    def test_valid_custom_imagespecfile(self, mock_super_validate):
        with TempDirectory() as test_dir:
            # setup
            parent_work_dir_path = os.path.join(test_dir.path, 'working')
            step_config = {
                'context': test_dir.path,
                'organization': 'mock-org',
                'service-name': 'mock-service',
                'application-name': 'mock-app',
                'imagespecfile': 'MockContainerfile.ubi8'
            }
            test_dir.write('MockContainerfile.ubi8',b'''testing''')

            # run test
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )
            step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

            # validate
            mock_super_validate.assert_called_once_with()

    def test_fail_missing_imagespecfile_defaults(self, mock_super_validate):
        with TempDirectory() as test_dir:
            # setup
            parent_work_dir_path = os.path.join(test_dir.path, 'working')
            step_config = {
                'context': test_dir.path,
                'organization': 'mock-org',
                'service-name': 'mock-service',
                'application-name': 'mock-app'
            }

            # run test
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )
            with self.assertRaisesRegex(
                AssertionError,
                rf'Given imagespecfile \(Containerfile\) does not exist'
                rf' in given context \({test_dir.path}\).'
            ):
                step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

            # validate
            mock_super_validate.assert_called_once_with()

    def test_fail_missing_imagespecfile_custom_imagespecfile(self, mock_super_validate):
        with TempDirectory() as test_dir:
            # setup
            parent_work_dir_path = os.path.join(test_dir.path, 'working')
            step_config = {
                'context': test_dir.path,
                'organization': 'mock-org',
                'service-name': 'mock-service',
                'application-name': 'mock-app',
                'imagespecfile': 'MockContainerfile.ubi8'
            }

            # run test
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )
            with self.assertRaisesRegex(
                AssertionError,
                rf'Given imagespecfile \(MockContainerfile.ubi8\) does not exist'
                rf' in given context \({test_dir.path}\).'
            ):
                step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

            # validate
            mock_super_validate.assert_called_once_with()

class TestStepImplementerCreateContainerImageBuildah_step_implementer_config_defaults(
    BaseTestStepImplementerCreateContainerImageBuildah
):
    def test_result(self):
        defaults = Buildah.step_implementer_config_defaults()
        expected_defaults = {
            'imagespecfile': 'Containerfile',
            'context': '.',
            'tls-verify': True,
            'format': 'oci'
        }
        self.assertEqual(defaults, expected_defaults)

class TestStepImplementerCreateContainerImageBuildah___required_config_or_result_keys(
    BaseTestStepImplementerCreateContainerImageBuildah
):
    def test_result(self):
        required_keys = Buildah._required_config_or_result_keys()
        expected_required_keys = [
            'imagespecfile',
            'context',
            'tls-verify',
            'format',
            'organization',
            'service-name',
            'application-name'
        ]
        self.assertEqual(required_keys, expected_required_keys)

@patch('ploigos_step_runner.step_implementers.create_container_image.buildah.determine_container_image_build_tag_info')
@patch('sh.buildah', create=True)
class TestStepImplementerCreateContainerImageBuildah___run_step(
    BaseTestStepImplementerCreateContainerImageBuildah
):
    def test_pass(
        self,
        buildah_mock,
        mock_determine_container_image_build_tag_info
    ):
        with TempDirectory() as temp_dir:
            # setup test
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            temp_dir.write('Containerfile',b'''testing''')

            artifact_config = {
                'container-image-version': {'description': '', 'value': '1.0-123abc'},
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {
                'imagespecfile': 'Containerfile',
                'context': temp_dir.path,
                'tls-verify': True,
                'format': 'oci',
                'organization': 'mock-org',
                'service-name': 'mock-service',
                'application-name': 'mock-app'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='create-container-image',
                implementer='Buildah',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            # set up mocks
            mock_determine_container_image_build_tag_info.return_value=[
                'localhost/mock-org/mock-app-mock-service:1.0-123abc',
                'mock-org/mock-app-mock-service:1.0-123abc',
                'localhost',
                'mock-app-mock-service',
                '1.0-123abc'
            ]

            # run test
            result = step_implementer._run_step()

            # verify results
            expected_step_result = StepResult(
                step_name='create-container-image',
                sub_step_name='Buildah',
                sub_step_implementer_name='Buildah'
            )
            expected_step_result.add_artifact(
                name='container-image-registry-uri',
                value='localhost',
                description='Registry URI poriton of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-registry-organization',
                value='mock-org',
                description='Organization portion of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-repository',
                value='mock-app-mock-service',
                description='Repository portion of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-name',
                value='mock-app-mock-service',
                description='Another way to reference the' \
                    ' repository portion of the container image tag of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-version',
                value='1.0-123abc',
                description='Version portion of the container image tag of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-tag',
                value='localhost/mock-org/mock-app-mock-service:1.0-123abc',
                description='Full container image tag of the built container,' \
                    ' including the registry URI.'
            )
            expected_step_result.add_artifact(
                name='container-image-short-tag',
                value='mock-org/mock-app-mock-service:1.0-123abc',
                description='Short container image tag of the built container image,' \
                    ' excluding the registry URI.'
            )
            self.assertEqual(result, expected_step_result)

            buildah_mock.bud.assert_called_once_with(
                '--format=oci',
                '--tls-verify=true',
                '--layers', '-f', 'Containerfile',
                '-t', 'localhost/mock-org/mock-app-mock-service:1.0-123abc',
                '--authfile', os.path.join(step_implementer.work_dir_path, 'container-auth.json'),
                temp_dir.path,
                _out=sys.stdout,
                _err=sys.stderr,
                _tee='err'
            )

    def test_pass_custom_auth_file(self, buildah_mock, mock_determine_container_image_build_tag_info):
        with TempDirectory() as temp_dir:
            # setup test
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            temp_dir.write('Containerfile',b'''testing''')

            artifact_config = {
                'container-image-version': {'description': '', 'value': '1.0-123abc'},
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {
                'imagespecfile': 'Containerfile',
                'context': temp_dir.path,
                'tls-verify': True,
                'format': 'oci',
                'organization': 'mock-org',
                'service-name': 'mock-service',
                'application-name': 'mock-app',
                'containers-config-auth-file': 'mock-auth.json'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='create-container-image',
                implementer='Buildah',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            # set up mocks
            mock_determine_container_image_build_tag_info.return_value=[
                'localhost/mock-org/mock-app-mock-service:1.0-123abc',
                'mock-org/mock-app-mock-service:1.0-123abc',
                'localhost',
                'mock-app-mock-service',
                '1.0-123abc'
            ]

            # run test
            result = step_implementer._run_step()

            # verify results
            expected_step_result = StepResult(
                step_name='create-container-image',
                sub_step_name='Buildah',
                sub_step_implementer_name='Buildah'
            )
            expected_step_result.add_artifact(
                name='container-image-registry-uri',
                value='localhost',
                description='Registry URI poriton of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-registry-organization',
                value='mock-org',
                description='Organization portion of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-repository',
                value='mock-app-mock-service',
                description='Repository portion of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-name',
                value='mock-app-mock-service',
                description='Another way to reference the' \
                    ' repository portion of the container image tag of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-version',
                value='1.0-123abc',
                description='Version portion of the container image tag of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-tag',
                value='localhost/mock-org/mock-app-mock-service:1.0-123abc',
                description='Full container image tag of the built container,' \
                    ' including the registry URI.'
            )
            expected_step_result.add_artifact(
                name='container-image-short-tag',
                value='mock-org/mock-app-mock-service:1.0-123abc',
                description='Short container image tag of the built container image,' \
                    ' excluding the registry URI.'
            )
            self.assertEqual(result, expected_step_result)

            buildah_mock.bud.assert_called_once_with(
                '--format=oci',
                '--tls-verify=true',
                '--layers', '-f', 'Containerfile',
                '-t', 'localhost/mock-org/mock-app-mock-service:1.0-123abc',
                '--authfile', 'mock-auth.json',
                temp_dir.path,
                _out=sys.stdout,
                _err=sys.stderr,
                _tee='err'
            )

    def test_pass_string_tls_verify(self, buildah_mock, mock_determine_container_image_build_tag_info):
        with TempDirectory() as temp_dir:
            # setup test
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            temp_dir.write('Containerfile',b'''testing''')

            artifact_config = {
                'container-image-version': {'description': '', 'value': '1.0-123abc'},
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {
                'imagespecfile': 'Containerfile',
                'context': temp_dir.path,
                'tls-verify': 'true',
                'format': 'oci',
                'organization': 'mock-org',
                'service-name': 'mock-service',
                'application-name': 'mock-app'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='create-container-image',
                implementer='Buildah',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            # setup sideeffects
            mock_determine_container_image_build_tag_info.return_value=[
                'localhost/mock-org/mock-app-mock-service:1.0-123abc',
                'mock-org/mock-app-mock-service:1.0-123abc',
                'localhost',
                'mock-app-mock-service',
                '1.0-123abc'
            ]

            # run test
            result = step_implementer._run_step()

            # verify results
            expected_step_result = StepResult(
                step_name='create-container-image',
                sub_step_name='Buildah',
                sub_step_implementer_name='Buildah'
            )
            expected_step_result.add_artifact(
                name='container-image-registry-uri',
                value='localhost',
                description='Registry URI poriton of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-registry-organization',
                value='mock-org',
                description='Organization portion of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-repository',
                value='mock-app-mock-service',
                description='Repository portion of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-name',
                value='mock-app-mock-service',
                description='Another way to reference the' \
                    ' repository portion of the container image tag of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-version',
                value='1.0-123abc',
                description='Version portion of the container image tag of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-tag',
                value='localhost/mock-org/mock-app-mock-service:1.0-123abc',
                description='Full container image tag of the built container,' \
                    ' including the registry URI.'
            )
            expected_step_result.add_artifact(
                name='container-image-short-tag',
                value='mock-org/mock-app-mock-service:1.0-123abc',
                description='Short container image tag of the built container image,' \
                    ' excluding the registry URI.'
            )
            self.assertEqual(result, expected_step_result)

            buildah_mock.bud.assert_called_once_with(
                '--format=oci',
                '--tls-verify=true',
                '--layers', '-f', 'Containerfile',
                '-t', 'localhost/mock-org/mock-app-mock-service:1.0-123abc',
                '--authfile', os.path.join(step_implementer.work_dir_path, 'container-auth.json'),
                temp_dir.path,
                _out=sys.stdout,
                _err=sys.stderr,
                _tee='err'
            )

    def test_pass_no_container_image_version(self, buildah_mock, mock_determine_container_image_build_tag_info):
        with TempDirectory() as temp_dir:
            # setup test
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            temp_dir.write('Containerfile',b'''testing''')
            step_config = {
                'imagespecfile': 'Containerfile',
                'context': temp_dir.path,
                'tls-verify': True,
                'format': 'oci',
                'organization': 'mock-org',
                'service-name': 'mock-service',
                'application-name': 'mock-app'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='create-container-image',
                implementer='Buildah',
                parent_work_dir_path=parent_work_dir_path,
            )

            # setup sideeffects
            mock_determine_container_image_build_tag_info.return_value=[
                'localhost/mock-org/mock-app-mock-service:mock-default-version',
                'mock-org/mock-app-mock-service:mock-default-version',
                'localhost',
                'mock-app-mock-service',
                'mock-default-version'
            ]

            # run test
            result = step_implementer._run_step()

            # verify results
            expected_step_result = StepResult(
                step_name='create-container-image',
                sub_step_name='Buildah',
                sub_step_implementer_name='Buildah'
            )
            expected_step_result.add_artifact(
                name='container-image-registry-uri',
                value='localhost',
                description='Registry URI poriton of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-registry-organization',
                value='mock-org',
                description='Organization portion of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-repository',
                value='mock-app-mock-service',
                description='Repository portion of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-name',
                value='mock-app-mock-service',
                description='Another way to reference the' \
                    ' repository portion of the container image tag of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-version',
                value='mock-default-version',
                description='Version portion of the container image tag of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-tag',
                value='localhost/mock-org/mock-app-mock-service:mock-default-version',
                description='Full container image tag of the built container,' \
                    ' including the registry URI.'
            )
            expected_step_result.add_artifact(
                name='container-image-short-tag',
                value='mock-org/mock-app-mock-service:mock-default-version',
                description='Short container image tag of the built container image,' \
                    ' excluding the registry URI.'
            )
            self.assertEqual(result, expected_step_result)

            buildah_mock.bud.assert_called_once_with(
                '--format=oci',
                '--tls-verify=true',
                '--layers', '-f', 'Containerfile',
                '-t', 'localhost/mock-org/mock-app-mock-service:mock-default-version',
                '--authfile', os.path.join(step_implementer.work_dir_path, 'container-auth.json'),
                temp_dir.path,
                _out=sys.stdout,
                _err=sys.stderr,
                _tee='err'
            )

    def test_fail_buildah_bud_error(self, buildah_mock, mock_determine_container_image_build_tag_info):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            temp_dir.write('Containerfile',b'''testing''')

            artifact_config = {
                'container-image-version': {'description': '', 'value': '1.0-123abc'},
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            image_spec_file = 'Containerfile'
            step_config = {
                'imagespecfile': image_spec_file,
                'context': temp_dir.path,
                'tls-verify': True,
                'format': 'oci',
                'service-name': 'mock-service',
                'application-name': 'mock-app'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='create-container-image',
                implementer='Buildah',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            # setup sideeffects
            mock_determine_container_image_build_tag_info.return_value=[
                'localhost/mock-org/mock-app-mock-service:1.0-123abc',
                'mock-org/mock-app-mock-service:1.0-123abc',
                'localhost',
                'mock-app-mock-service',
                '1.0-123abc'
            ]

            # run step with mock failure
            buildah_mock.bud.side_effect = sh.ErrorReturnCode('buildah', b'mock out', b'mock error')
            result = step_implementer._run_step()

             # verify step result
            self.assertFalse(result.success)
            self.assertRegex(
                result.message,
                re.compile(
                    r'Issue invoking buildah bud with given image '
                    rf'specification file \({image_spec_file}\):'
                    r'.*RAN: buildah'
                    r'.*STDOUT:'
                    r'.*mock out'
                    r'.*STDERR:'
                    r'.*mock error',
                    re.DOTALL
                )
            )



"""Test ResultStep
"""
from ploigos_step_runner import StepResult, WorkflowResult
from ploigos_step_runner.config import Config
from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner.results import (StepResultArtifact,
                                         StepResultEvidence, step_result)
from tests.helpers.base_test_case import BaseTestCase
from tests.helpers.sample_step_implementers import FooStepImplementer


class TestStepResultTest(BaseTestCase):
    """Test StepResult
    """

    def test_step_name(self):
        step_result_expected = 'step1'
        step_result = StepResult('step1', 'sub1', 'implementer1')
        self.assertEqual(step_result.step_name, step_result_expected)

    def test_sub_step_name(self):
        expected = 'sub1'
        step_result = StepResult('step1', 'sub1', 'implementer1')
        self.assertEqual(step_result.sub_step_name, expected)

    def test_sub_step_implementer_name(self):
        expected = 'implementer1'
        step_result = StepResult('step1', 'sub1', 'implementer1')
        self.assertEqual(step_result.sub_step_implementer_name, expected)

    def test_success(self):
        step_result = StepResult('step1', 'sub1', 'implementer1')
        step_result.success = False
        self.assertEqual(step_result.success, False)

    def test_message(self):
        step_result = StepResult('step1', 'sub1', 'implementer1')
        step_result.message = 'testing'
        self.assertEqual(step_result.message, 'testing')

    def test_add_artifact(self):
        step_result = StepResult('step1', 'sub1', 'implementer1')
        step_result.add_artifact('artifact1', 'value1', 'description1')
        step_result.add_artifact('artifact2', 'value2', 'description2')
        step_result.add_artifact('artifact3', 'value3')

        self.assertEqual(
            step_result.get_artifact('artifact1'),
            StepResultArtifact(
                name='artifact1',
                value='value1',
                description='description1'
            )
        )
        self.assertEqual(
            step_result.get_artifact('artifact2'),
            StepResultArtifact(
                name='artifact2',
                value='value2',
                description='description2'
            )
        )
        self.assertEqual(
            step_result.get_artifact('artifact3'),
            StepResultArtifact(
                name='artifact3',
                value='value3'
            )
        )

    def test_add_evidence(self):
        step_result = StepResult('step1', 'sub1', 'implementer1')
        step_result.add_evidence('evidence1', 'value1', 'description1')
        step_result.add_evidence('evidence2', 'value2', 'description2')
        step_result.add_evidence('evidence3', 'value3')

        self.assertEqual(
            step_result.get_evidence('evidence1'),
            StepResultEvidence(
                name='evidence1',
                value='value1',
                description='description1'
            )
        )
        self.assertEqual(
            step_result.get_evidence('evidence2'),
            StepResultEvidence(
                name='evidence2',
                value='value2',
                description='description2'
            )
        )
        self.assertEqual(
            step_result.get_evidence('evidence3'),
            StepResultEvidence(
                name='evidence3',
                value='value3'
            )
        )

    def test_add_artifact_missing_name(self):
        step_result = StepResult('step1', 'sub1', 'implementer1')

        with self.assertRaisesRegex(
                StepRunnerException,
                r"Name is required to add artifact"):
            step_result.add_artifact('', 'value1', 'description1')

    def test_add_evidence_missing_name(self):
        step_result = StepResult('step1', 'sub1', 'implementer1')

        with self.assertRaisesRegex(
                StepRunnerException,
                r"Name is required to add evidence"):
            step_result.add_evidence('', 'value1', 'description1')

    def test_add_artifact_missing_value(self):
        step_result = StepResult('step1', 'sub1', 'implementer1')

        with self.assertRaisesRegex(
                StepRunnerException,
                r"Value is required to add artifact"):
            step_result.add_artifact('name1', '', 'description1')

    def test_add_evidence_missing_value(self):
        step_result = StepResult('step1', 'sub1', 'implementer1')

        with self.assertRaisesRegex(
                StepRunnerException,
                r"Value is required to add evidence"):
            step_result.add_evidence('name1', '', 'description1')

    def test_get_artifact(self):
        expected_artifact = StepResultArtifact(
            name='artifact1',
            value='value1',
            description='description1'
        )

        step_result = StepResult('step1', 'sub1', 'implementer1')
        step_result.add_artifact('artifact1', 'value1', 'description1')
        self.assertEqual(step_result.get_artifact('artifact1'), expected_artifact)

    def test_get_evidence(self):
        expected_evidence = StepResultEvidence(
            name='evidence1',
            value='value1',
            description='description1'
        )

        step_result = StepResult('step1', 'sub1', 'implementer1')
        step_result.add_evidence('evidence1', 'value1', 'description1')
        self.assertEqual(step_result.get_evidence('evidence1'), expected_evidence)

    def test_get_artifact_value(self):
        step_result_expected = 'value1'
        step_result = StepResult('step1', 'sub1', 'implementer1')
        step_result.add_artifact('artifact1', 'value1', 'description1')
        self.assertEqual(step_result.get_artifact_value('artifact1'), step_result_expected)

    def test_get_evidence_value(self):
        step_result_expected = 'value1'
        step_result = StepResult('step1', 'sub1', 'implementer1')
        step_result.add_evidence('evidence1', 'value1', 'description1')
        self.assertEqual(step_result.get_evidence_value('evidence1'), step_result_expected)

    def test_get_artifacts_property(self):
        step_result = StepResult('step1', 'sub1', 'implementer1')
        step_result.add_artifact('artifact1', 'value1', 'description1')
        step_result.add_artifact('artifact2', 'value2')

        expected_artifacts = {
            'artifact1': StepResultArtifact(name='artifact1', value='value1', description='description1'),
            'artifact2': StepResultArtifact(name='artifact2', value='value2')
        }

        self.assertEqual(step_result.artifacts, expected_artifacts)

    def test_get_evidence_property(self):
        step_result = StepResult('step1', 'sub1', 'implementer1')
        step_result.add_evidence('evidence1', 'value1', 'description1')
        step_result.add_evidence('evidence2', 'value2')

        expected_evidence = {
            'evidence1': StepResultEvidence(name='evidence1', value='value1', description='description1'),
            'evidence2': StepResultEvidence(name='evidence2', value='value2')
        }

        self.assertEqual(step_result.evidence, expected_evidence)

    def test_add_duplicate_artifact(self):
        expected_artifacts = {
            'artifact1': StepResultArtifact(name='artifact1', value='value1', description='description1'),
            'artifact2': StepResultArtifact(name='artifact2', value='lastonewins')
        }

        step_result = StepResult('step1', 'sub1', 'implementer1')
        step_result.add_artifact('artifact1', 'value1', 'description1')
        step_result.add_artifact('artifact2', 'here')
        step_result.add_artifact('artifact2', 'andhere')
        step_result.add_artifact('artifact2', 'lastonewins')
        self.assertEqual(step_result.artifacts, expected_artifacts)

    def test_add_duplicate_evidence(self):
        expected_evidence = {
            'evidence1': StepResultEvidence(name='evidence1', value='value1', description='description1'),
            'evidence2': StepResultEvidence(name='evidence2', value='lastonewins')
        }

        step_result = StepResult('step1', 'sub1', 'implementer1')
        step_result.add_evidence('evidence1', 'value1', 'description1')
        step_result.add_evidence('evidence2', 'here')
        step_result.add_evidence('evidence2', 'andhere')
        step_result.add_evidence('evidence2', 'lastonewins')
        self.assertEqual(step_result.evidence, expected_evidence)

    def test_from_step_implementer_no_env(self):
        config = Config({
            'step-runner-config': {
                'foo': {
                    'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer',
                    'config': {}
                }
            }
        })
        step_config = config.get_step_config('foo')
        sub_step = step_config.get_sub_step(
            'tests.helpers.sample_step_implementers.FooStepImplementer')

        step = FooStepImplementer(
            workflow_result=WorkflowResult(),
            parent_work_dir_path=None,
            config=sub_step
        )

        step_result = StepResult.from_step_implementer(step)

        expected_step_result = StepResult(
            step_name='foo',
            sub_step_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer'
        )

        self.assertEqual(step_result, expected_step_result)

    def test_from_step_implementer_with_env(self):
        config = Config({
            'step-runner-config': {
                'foo': {
                    'implementer': 'tests.helpers.sample_step_implementers.FooStepImplementer',
                    'config': {}
                }
            }
        })
        step_config = config.get_step_config('foo')
        sub_step = step_config.get_sub_step(
            'tests.helpers.sample_step_implementers.FooStepImplementer')

        step = FooStepImplementer(
            workflow_result=WorkflowResult(),
            parent_work_dir_path=None,
            config=sub_step,
            environment='blarg'
        )

        step_result = StepResult.from_step_implementer(step)

        expected_step_result = StepResult(
            step_name='foo',
            sub_step_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            environment='blarg'
        )

        self.assertEqual(step_result, expected_step_result)

    def test_artifacts_dicts_empty(self):
        step_result = StepResult(
            step_name='foo',
            sub_step_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer'
        )

        self.assertEqual([], step_result.artifacts_dicts)

    def test_evidence_dicts_empty(self):
        step_result = StepResult(
            step_name='foo',
            sub_step_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer'
        )

        self.assertEqual([], step_result.evidence_dicts)

    def test_artifacts_dicts_with_artifacts(self):
        step_result = StepResult(
            step_name='foo',
            sub_step_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer'
        )
        step_result.add_artifact(
            name='art-str',
            value='hello'
        )
        step_result.add_artifact(
            name='art-bool-t',
            value=True
        )
        step_result.add_artifact(
            name='art-bool-f',
            value=False
        )
        step_result.add_artifact(
            name='art-desc',
            value='world',
            description='test artifact'
        )

        expected_step_result_artifacts_dicts = [
            {'description': '', 'name': 'art-str', 'value': 'hello'},
            {'description': '', 'name': 'art-bool-t', 'value': True},
            {'description': '', 'name': 'art-bool-f', 'value': False},
            {'description': 'test artifact', 'name': 'art-desc', 'value': 'world'}
        ]

        self.assertEqual(expected_step_result_artifacts_dicts, step_result.artifacts_dicts)

    def test_evidence_dicts_with_evidence(self):
        step_result = StepResult(
            step_name='foo',
            sub_step_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer'
        )
        step_result.add_evidence(
            name='evi-str',
            value='hello'
        )
        step_result.add_evidence(
            name='evi-bool-t',
            value=True
        )
        step_result.add_evidence(
            name='evi-bool-f',
            value=False
        )
        step_result.add_evidence(
            name='evi-desc',
            value='world',
            description='test evidence'
        )

        expected_step_result_evidence_dicts = [
            {'description': '', 'name': 'evi-str', 'value': 'hello'},
            {'description': '', 'name': 'evi-bool-t', 'value': True},
            {'description': '', 'name': 'evi-bool-f', 'value': False},
            {'description': 'test evidence', 'name': 'evi-desc', 'value': 'world'}
        ]

        self.assertEqual(expected_step_result_evidence_dicts, step_result.evidence_dicts)

    def test_get_sub_step_result_dict(self):
        step_result = StepResult(
            step_name='foo',
            sub_step_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer'
        )
        step_result.add_artifact(
            name='art-str',
            value='hello'
        )
        step_result.add_artifact(
            name='art-bool-t',
            value=True
        )
        step_result.add_artifact(
            name='art-bool-f',
            value=False
        )
        step_result.add_artifact(
            name='art-desc',
            value='world',
            description='test artifact'
        )

        step_result.add_evidence(
            name='evi-str',
            value='hello'
        )
        step_result.add_evidence(
            name='evi-bool-t',
            value=True
        )
        step_result.add_evidence(
            name='evi-bool-f',
            value=False
        )
        step_result.add_evidence(
            name='evi-desc',
            value='world',
            description='test evidence'
        )

        expected = {
            'artifacts': [
                {'description': '', 'name': 'art-str', 'value': 'hello'},
                {'description': '', 'name': 'art-bool-t', 'value': True},
                {'description': '', 'name': 'art-bool-f', 'value': False},
                {'description': 'test artifact', 'name': 'art-desc', 'value': 'world'}
            ],
            'evidence': [
                {'description': '', 'name': 'evi-str', 'value': 'hello'},
                {'description': '', 'name': 'evi-bool-t', 'value': True},
                {'description': '', 'name': 'evi-bool-f', 'value': False},
                {'description': 'test evidence', 'name': 'evi-desc', 'value': 'world'}
            ],
            'message': '',
            'sub-step-implementer-name': 'tests.helpers.sample_step_implementers.FooStepImplementer',
            'success': True
        }

        self.assertEqual(expected, step_result.get_sub_step_result_dict())

    def test_get_step_result_dict_no_env(self):
        step_result = StepResult(
            step_name='foo',
            sub_step_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer'
        )
        step_result.add_artifact(
            name='art-str',
            value='hello'
        )
        step_result.add_artifact(
            name='art-bool-t',
            value=True
        )
        step_result.add_artifact(
            name='art-bool-f',
            value=False
        )
        step_result.add_artifact(
            name='art-desc',
            value='world',
            description='test artifact'
        )

        step_result.add_evidence(
            name='evi-str',
            value='hello'
        )
        step_result.add_evidence(
            name='evi-bool-t',
            value=True
        )
        step_result.add_evidence(
            name='evi-bool-f',
            value=False
        )
        step_result.add_evidence(
            name='evi-desc',
            value='world',
            description='test evidence'
        )

        expected = {
            'foo': {
                'tests.helpers.sample_step_implementers.FooStepImplementer': {
                    'artifacts': [
                        {'description': '', 'name': 'art-str', 'value': 'hello'},
                        {'description': '', 'name': 'art-bool-t', 'value': True},
                        {'description': '', 'name': 'art-bool-f', 'value': False},
                        {'description': 'test artifact', 'name': 'art-desc', 'value': 'world'}
                    ],
                    'evidence': [
                        {'description': '', 'name': 'evi-str', 'value': 'hello'},
                        {'description': '', 'name': 'evi-bool-t', 'value': True},
                        {'description': '', 'name': 'evi-bool-f', 'value': False},
                        {'description': 'test evidence', 'name': 'evi-desc', 'value': 'world'}
                    ],
                    'message': '',
                    'sub-step-implementer-name': 'tests.helpers.sample_step_implementers.FooStepImplementer',
                    'success': True
                }
            }
        }

        self.assertEqual(expected, step_result.get_step_result_dict())

    def test_get_step_result_dict_with_env(self):
        step_result = StepResult(
            step_name='foo',
            sub_step_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            environment='blarg'
        )
        step_result.add_artifact(
            name='art-str',
            value='hello'
        )
        step_result.add_artifact(
            name='art-bool-t',
            value=True
        )
        step_result.add_artifact(
            name='art-bool-f',
            value=False
        )
        step_result.add_artifact(
            name='art-desc',
            value='world',
            description='test artifact'
        )

        step_result.add_evidence(
            name='evi-str',
            value='hello'
        )
        step_result.add_evidence(
            name='evi-bool-t',
            value=True
        )
        step_result.add_evidence(
            name='evi-bool-f',
            value=False
        )
        step_result.add_evidence(
            name='evi-desc',
            value='world',
            description='test evidence'
        )

        expected = {
            'blarg': {
                'foo': {
                    'tests.helpers.sample_step_implementers.FooStepImplementer': {
                        'artifacts': [
                            {'description': '', 'name': 'art-str', 'value': 'hello'},
                            {'description': '', 'name': 'art-bool-t', 'value': True},
                            {'description': '', 'name': 'art-bool-f', 'value': False},
                            {'description': 'test artifact', 'name': 'art-desc', 'value': 'world'}
                        ],
                        'evidence': [
                            {'description': '', 'name': 'evi-str', 'value': 'hello'},
                            {'description': '', 'name': 'evi-bool-t', 'value': True},
                            {'description': '', 'name': 'evi-bool-f', 'value': False},
                            {'description': 'test evidence', 'name': 'evi-desc', 'value': 'world'}
                        ],
                        'message': '',
                        'sub-step-implementer-name': 'tests.helpers.sample_step_implementers.FooStepImplementer',
                        'success': True
                    }
                }
            }
        }

        self.assertEqual(expected, step_result.get_step_result_dict())

    def test___str__(self):
        step_result = StepResult(
            step_name='foo',
            sub_step_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            environment='blarg'
        )
        step_result.add_artifact(
            name='art-str',
            value='hello'
        )
        step_result.add_artifact(
            name='art-bool-t',
            value=True
        )
        step_result.add_artifact(
            name='art-bool-f',
            value=False
        )
        step_result.add_artifact(
            name='art-desc',
            value='world',
            description='test artifact'
        )

        step_result.add_evidence(
            name='evi-str',
            value='hello'
        )
        step_result.add_evidence(
            name='evi-bool-t',
            value=True
        )
        step_result.add_evidence(
            name='evi-bool-f',
            value=False
        )
        step_result.add_evidence(
            name='evi-desc',
            value='world',
            description='test evidence'
        )

        expected = "{'step-name': 'foo', 'sub-step-name': "\
        "'tests.helpers.sample_step_implementers.FooStepImplementer', "\
        "'sub-step-implementer-name': 'tests.helpers.sample_step_implementers.FooStepImplementer', "\
        "'environment': 'blarg', 'success': True, 'message': '', "\
        "'artifacts': [{'name': 'art-str', 'value': 'hello', 'description': ''}, "\
        "{'name': 'art-bool-t', 'value': True, 'description': ''}, "\
        "{'name': 'art-bool-f', 'value': False, 'description': ''}, "\
        "{'name': 'art-desc', 'value': 'world', 'description': 'test artifact'}], "\
        "'evidence': [{'name': 'evi-str', 'value': 'hello', 'description': ''}, "\
        "{'name': 'evi-bool-t', 'value': True, 'description': ''}, "\
        "{'name': 'evi-bool-f', 'value': False, 'description': ''}, "\
        "{'name': 'evi-desc', 'value': 'world', 'description': 'test evidence'}]}"

        self.assertEqual(expected, str(step_result))

    def test___repr__(self):
        step_result = StepResult(
            step_name='foo',
            sub_step_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            environment='blarg'
        )
        step_result.add_artifact(
            name='art-str',
            value='hello'
        )
        step_result.add_artifact(
            name='art-bool-t',
            value=True
        )
        step_result.add_artifact(
            name='art-bool-f',
            value=False
        )
        step_result.add_artifact(
            name='art-desc',
            value='world',
            description='test artifact'
        )

        step_result.add_evidence(
            name='evi-str',
            value='hello'
        )
        step_result.add_evidence(
            name='evi-bool-t',
            value=True
        )
        step_result.add_evidence(
            name='evi-bool-f',
            value=False
        )
        step_result.add_evidence(
            name='evi-desc',
            value='world',
            description='test evidence'
        )

        expected = "StepResult(step_name=foo,sub_step_name=tests.helpers." \
            "sample_step_implementers.FooStepImplementer,sub_step_implementer_name=" \
            "tests.helpers.sample_step_implementers.FooStepImplementer,environment=" \
            "blarg,success=True,message=,artifacts=[{'name': 'art-str', 'value': 'hello', "\
            "'description': ''}, {'name': 'art-bool-t', 'value': True, 'description': ''}, "\
            "{'name': 'art-bool-f', 'value': False, 'description': ''}, "\
            "{'name': 'art-desc', 'value': 'world', 'description': 'test artifact'}]"\
            "evidence=[{'name': 'evi-str', 'value': 'hello', 'description': ''}, "\
            "{'name': 'evi-bool-t', 'value': True, 'description': ''}, "\
            "{'name': 'evi-bool-f', 'value': False, 'description': ''}, "\
            "{'name': 'evi-desc', 'value': 'world', 'description': 'test evidence'}])"


        print(str(repr(step_result)))

        self.assertEqual(expected, repr(step_result))

    def test___eq__(self):
        step_result1 = StepResult(
            step_name='foo',
            sub_step_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            environment='blarg'
        )
        step_result1.add_artifact(
            name='art-str',
            value='hello'
        )
        step_result1.add_artifact(
            name='art-bool-t',
            value=True
        )
        step_result1.add_artifact(
            name='art-bool-f',
            value=False
        )
        step_result1.add_artifact(
            name='art-desc',
            value='world',
            description='test artifact'
        )

        step_result1.add_evidence(
            name='evi-str',
            value='hello'
        )
        step_result1.add_evidence(
            name='evi-bool-t',
            value=True
        )
        step_result1.add_evidence(
            name='evi-bool-f',
            value=False
        )
        step_result1.add_evidence(
            name='evi-desc',
            value='world',
            description='test evidence'
        )

        step_result2 = StepResult(
            step_name='foo',
            sub_step_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            environment='blarg'
        )
        step_result2.add_artifact(
            name='art-str',
            value='hello'
        )
        step_result2.add_artifact(
            name='art-bool-t',
            value=True
        )
        step_result2.add_artifact(
            name='art-bool-f',
            value=False
        )
        step_result2.add_artifact(
            name='art-desc',
            value='world',
            description='test artifact'
        )

        step_result2.add_evidence(
            name='evi-str',
            value='hello'
        )
        step_result2.add_evidence(
            name='evi-bool-t',
            value=True
        )
        step_result2.add_evidence(
            name='evi-bool-f',
            value=False
        )
        step_result2.add_evidence(
            name='evi-desc',
            value='world',
            description='test evidence'
        )

        self.assertEqual(step_result1, step_result2)

    def test___nq___artifacts(self):
        step_result1 = StepResult(
            step_name='foo',
            sub_step_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            environment='blarg'
        )
        step_result1.add_artifact(
            name='art-str',
            value='hello'
        )
        step_result1.add_artifact(
            name='art-bool-t',
            value=True
        )
        step_result1.add_artifact(
            name='art-bool-f',
            value=False
        )
        step_result1.add_artifact(
            name='art-desc',
            value='world',
            description='test artifact'
        )

        step_result2 = StepResult(
            step_name='foo',
            sub_step_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            environment='blarg'
        )

        self.assertNotEqual(step_result1, step_result2)

    def test___nq___step_name(self):
        step_result1 = StepResult(
            step_name='foo',
            sub_step_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            environment='blarg'
        )

        step_result2 = StepResult(
            step_name='foo1',
            sub_step_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            environment='blarg'
        )

        self.assertNotEqual(step_result1, step_result2)

    def test___nq___sub_step_name(self):
        step_result1 = StepResult(
            step_name='foo',
            sub_step_name='a',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            environment='blarg'
        )

        step_result2 = StepResult(
            step_name='foo',
            sub_step_name='b',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            environment='blarg'
        )

        self.assertNotEqual(step_result1, step_result2)

    def test___nq___sub_step_implementer_name(self):
        step_result1 = StepResult(
            step_name='foo',
            sub_step_name='a',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            environment='blarg'
        )

        step_result2 = StepResult(
            step_name='foo',
            sub_step_name='a',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.BarStepImplementer',
            environment='blarg'
        )

        self.assertNotEqual(step_result1, step_result2)

    def test___nq___environment(self):
        step_result1 = StepResult(
            step_name='foo',
            sub_step_name='a',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            environment='a'
        )

        step_result2 = StepResult(
            step_name='foo',
            sub_step_name='a',
            sub_step_implementer_name='tests.helpers.sample_step_implementers.FooStepImplementer',
            environment='b'
        )

        self.assertNotEqual(step_result1, step_result2)

    def test_merge(self):
       sr1 = StepResult('step1', 'sub1', 'implementer1')
       sr1.add_artifact('artifact1', 'value1', 'description1')
       sr1.add_artifact('artifact2', 'value2', 'description2')
       sr1.add_evidence('evidence1', 'value1', 'description1')
       sr1.add_evidence('evidence2', 'value2', 'description2')

       sr2 = StepResult('step1', 'sub1', 'implementer1')
       sr2.add_artifact('artifact1', 'changed-value1', 'changed-description1')
       sr2.add_evidence('evidence1', 'changed-value1', 'changed-description1')

       sr1.merge(sr2)

       artifact1 = sr1.get_artifact('artifact1')
       artifact2 = sr1.get_artifact('artifact2')
       evidence1 = sr1.get_evidence('evidence1')
       evidence2 = sr1.get_evidence('evidence2')

        # changed by SR2
       self.assertEqual(artifact1.value, 'changed-value1')
       self.assertEqual(artifact1.description, 'changed-description1')
       self.assertEqual(evidence1.value, 'changed-value1')
       self.assertEqual(evidence1.description, 'changed-description1')

       # unchanged by SR2
       self.assertEqual(artifact2.value, 'value2')
       self.assertEqual(artifact2.description, 'description2')
       self.assertEqual(evidence2.value, 'value2')
       self.assertEqual(evidence2.description, 'description2')

    def test_merge_raises_exception_for_mismatching_data(self):
        sr1 = StepResult('step1', 'sub1', 'implementer1')
        sr2 = StepResult('step2', 'sub1', 'implementer1')

        with self.assertRaisesRegex(
                StepRunnerException,
                f'Other StepResult does not have matching ' \
                f'step name, sub step name, or environment'
            ):
            sr1.merge(sr2)

    def test_merge_raises_exception_on_non_stepresult_other(self):
        sr1 = StepResult('step1', 'sub1', 'implementer1')
        other = "string"

        with self.assertRaisesRegex(
                StepRunnerException,
                f'expect StepResult instance type'
             ):
            sr1.merge(other)



"""`StepImplementer` for the `generate-metadata` step using NPM.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:

  * static configuration
  * runtime configuration
  * previous step results

Configuration Key | Required? | Default          | Description
------------------|-----------|------------------|-----------
`package-file`    | True      | `'package.json'` | node.js package file to read \
                                                   the app version out of

Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key | Description
--------------------|------------
`app-version`       | Value to use for `version` portion of semantic version \
                      (https://semver.org/). Uses the version read out of the given pom file.
"""

import json
import os.path

from ploigos_step_runner import StepImplementer
from ploigos_step_runner import StepResult

DEFAULT_CONFIG = {
    'package-file': 'package.json'
}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'package-file'
]

class Npm(StepImplementer): # pylint: disable=too-few-public-methods
    """`StepImplementer` for the `generate-metadata` step using NPM.
    """

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Notes
        -----
        These are the lowest precedence configuration values.

        Returns
        -------
        dict
            Default values to use for step configuration values.
        """
        return DEFAULT_CONFIG

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _validate_required_config_or_previous_step_result_artifact_keys(self):
        """Validates that the required configuration keys or previous step result artifacts
        are set and have valid values.

        Validates that:
        * required configuration is given
        * given 'package-file' exists

        Raises
        ------
        AssertionError
            If step configuration or previous step result artifacts have invalid required values
        """
        super()._validate_required_config_or_previous_step_result_artifact_keys()

        package_file = self.get_value('package-file')
        assert os.path.exists(package_file), \
            f'Given npm package file (package-file) does not exist: {package_file}'

    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        package_file = self.get_value('package-file')

        with open(package_file, encoding='utf-8') as package_file_object:
            package_file_data = json.load(package_file_object)

        if not "version" in package_file_data:
            step_result.success = False
            step_result.message = f'Given npm package file ({package_file})' + \
              ' does not contain a \"version\" key.'
            return step_result

        step_result.add_artifact(
            name='app-version',
            value=package_file_data["version"]
        )

        return step_result



"""Step Implementer for the container-image-static-compliance-scan step for OpenSCAP.

Step Configuration
------------------

Step configuration expected as input to this step.
Could come from either configuration file or
from runtime configuration.

 Configuration Key             | Required? | Default | Description
-------------------------------|-----------|---------|-----------
`container-image-tag`          | Yes       |         | Container image tag to scan.
`oscap-input-definitions-uri`  | Yes       |         | URI to the OpenSCAP definitions file \
                                                       to do the evaluation with. \
                                                       Must use protocol file://|http://|https://. \
                                                       Must have file extension .xml|.bz2.
`oscap-profile`                | No        |         | OpenSCAP profile to evaluate.
`oscap-tailoring-uri`          | No        |         | URI to OpenSCAP tailoring file \
                                                       to do the evaluation with. \
                                                       Must use protocol file://|http://|https://. \
                                                       Must have file extension .xml|.bz2.
`oscap-fetch-remote-resources` | No        | True    | For Source DataStream and XCCDF files \
                                                       that have remote references fetch them if \
                                                       True, else don't. \
                                                       <br/><br/> \
                                                       **WARNING**: evaluations will not be \
                                                       complete if input defintions require \
                                                       remote resources and this is not True. \
                                                       For disconnected environments the remote \
                                                       internal mirror.
`[container-image-pull-repository-type, container-image-repository-type]` \
                               | Yes       | 'containers-storage:' \
                                                     | \
                                           Container repository type for the pull image source. \
                                           See https://github.com/containers/skopeo for valid \
                                           options.

Results
-------

Results output by this step.

| Result Key      | Description
|-----------------|------------
| `html-report`   | HTML report generated by oscap eval
| `xml-report`    | XML report generated by oscap eval
| `stdout-report` | stdout report generated by oscap eval
"""

import os
import re
from distutils.util import strtobool
from io import StringIO

import sh
from ploigos_step_runner import StepResult, StepRunnerException
from ploigos_step_runner.step_implementer import StepImplementer
from ploigos_step_runner.utils.containers import (create_container_from_image,
                                                  mount_container)
from ploigos_step_runner.utils.file import \
    download_and_decompress_source_to_destination
from ploigos_step_runner.utils.io import \
    create_sh_redirect_to_multiple_streams_fn_callback

DEFAULT_CONFIG = {
    'oscap-fetch-remote-resources': True,
    'container-image-pull-repository-type': 'containers-storage:',
    'container-image-repository-type': 'containers-storage:'
}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'oscap-input-definitions-uri',
    'container-image-tag',
    'container-image-pull-repository-type',

    # being flexible for different use cases of proceeding steps
    ['container-image-pull-repository-type', 'container-image-repository-type']
]


class OpenSCAPGeneric(StepImplementer):
    """A generic OpenSCAP step implementer that can be used for more then one step.

    Expected uses:
    * container-image-static-compliance-scan
    * container-image-static-vulnerability-scan
    """

    # Example Input:
    #    Title	RHSA-2020:4186: spice and spice-gtk security update (Important)
    #    Rule	xccdf_com.redhat.rhsa_rule_oval-com.redhat.rhsa-def-20204186
    #    Ident	RHSA-2020:4186
    #    Ident	CVE-2020-14355
    #    Result	pass
    #
    #    Title	RHSA-2020:3658: librepo security update (Important)
    #    Rule	xccdf_com.redhat.rhsa_rule_oval-com.redhat.rhsa-def-20203658
    #    Ident	RHSA-2020:3658
    #    Ident	CVE-2020-14352
    #    Result	fail
    #
    # Matches:
    #    (Title	RHSA-2020:4186: spice and spice-gtk security update (Important)
    #    Rule	xccdf_com.redhat.rhsa_rule_oval-com.redhat.rhsa-def-20204186
    #    Ident	RHSA-2020:4186
    #    Ident	CVE-2020-14355
    #    Result	(pass))
    #
    #    (Title	RHSA-2020:3658: librepo security update (Important)
    #    Rule	xccdf_com.redhat.rhsa_rule_oval-com.redhat.rhsa-def-20203658
    #    Ident	RHSA-2020:3658
    #    Ident	CVE-2020-14352
    #    Result	(fail))
    #
    # Named Groups:
    #    [0]ruleblock
    #        Title	RHSA-2020:4186: spice and spice-gtk security update (Important)
    #        Rule	xccdf_com.redhat.rhsa_rule_oval-com.redhat.rhsa-def-20204186
    #        Ident	RHSA-2020:4186
    #        Ident	CVE-2020-14355
    #        Result	pass
    #    [0]ruleresult
    #        pass
    #
    #    [1]ruleblock
    #        Title	RHSA-2020:3658: librepo security update (Important)
    #        Rule	xccdf_com.redhat.rhsa_rule_oval-com.redhat.rhsa-def-20203658
    #        Ident	RHSA-2020:3658
    #        Ident	CVE-2020-14352
    #        Result	fail
    #    [1]ruleresult
    #        fail
    OSCAP_XCCDF_STDOUT_PATTERN = re.compile(
        r'(?P<ruleblock>Title.+?Result\s+(?P<ruleresult>[^\n]+))\n',
        re.DOTALL
    )
    OSCAP_XCCDF_STDOUT_FAIL_PATTERN = re.compile(r'fail')

    # NOTE: oval output far less useful then xccdf output but it is all but given some content
    #       is only given in oval format and therefor supporting this is important
    #
    # Example Input:
    #   Definition oval:com.redhat.rhsa:def:20202031: false
    #   Definition oval:com.redhat.rhsa:def:20201998: true
    #
    # Matches:
    #   (Definition oval:com.redhat.rhsa:def:20202031: (false))
    #   (Definition oval:com.redhat.rhsa:def:20201998: (true))
    #
    # Named Groups:
    #   [0]ruleblock
    #       Definition oval:com.redhat.rhsa:def:20202031: false
    #   [0]ruleresult
    #       false
    #
    #   [1]ruleblock
    #       Definition oval:com.redhat.rhsa:def:20201998: true
    #   [1]ruleresult
    #       true
    OSCAP_OVAL_STDOUT_PATTERN = re.compile(
        r'(?P<ruleblock>^.*:\s*(?P<ruleresult>true|false)\s*$)$',
        re.MULTILINE
    )
    OSCAP_OVAL_STDOUT_FAIL_PATTERN = re.compile(r'true')

    OSCAP_INFO_DOC_TYPE_PATTERN = re.compile(r'Document type: (?P<doctype>.+)')

    @staticmethod
    def step_implementer_config_defaults():
        """
        Getter for the StepImplementer's configuration defaults.

        Notes
        -----
        These are the lowest precedence configuration values.

        Returns
        -------
        dict
            Default values to use for step configuration values.
        """
        return DEFAULT_CONFIG

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _validate_required_config_or_previous_step_result_artifact_keys(self):
        """Validates that the required configuration keys or previous step result artifacts
        are set and have valid values.

        Validates that:
        * required configuration is given
        * oscap-input-definitions-uri
          - starts with file://|http://|https://
          - ends with .xml|.bz2

        Raises
        ------
        AssertionError
            If step configuration or previous step result artifacts have invalid required values
        """
        super()._validate_required_config_or_previous_step_result_artifact_keys()  # pylint: disable=protected-access

        # validate that the given 'oscap-input-definitions-uri' starts with file://|http://|https://
        oscap_input_definitions_uri = self.get_value('oscap-input-definitions-uri')
        assert (re.match(r'^file://|http://|https://', oscap_input_definitions_uri)), \
            f"Open SCAP input definitions source ({oscap_input_definitions_uri})" \
            f" must start with known protocol (file://|http://|https://)."

        # validate that the given 'oscap-input-definitions-uri' is an xml or bz2 file
        oscap_input_definitions_uri_extension = os.path.splitext(oscap_input_definitions_uri)[1]
        assert (re.match(r'\.xml|\.bz2', oscap_input_definitions_uri_extension)), \
            f"Open SCAP input definitions source ({oscap_input_definitions_uri})" \
            f" must be of known type (xml|bz2), got: {oscap_input_definitions_uri_extension}"

    def _run_step(self):  # pylint: disable=too-many-locals,too-many-statements
        """Runs the OpenSCAP eval for a given input file against a given container.
        """
        step_result = StepResult.from_step_implementer(self)

        # get config
        image_tag = self.get_value('container-image-tag')
        oscap_profile = self.get_value('oscap-profile')
        oscap_fetch_remote_resources = self.get_value('oscap-fetch-remote-resources')
        pull_repository_type = self.get_value([
            'container-image-pull-repository-type',
            'container-image-repository-type'
        ])

        try:
            # create container from image that can be mounted
            print(f"\nCreate container from image ({image_tag})")
            container_name = create_container_from_image(
                image_tag=image_tag,
                repository_type=pull_repository_type
            )
            print(f"Created container ({container_name}) from image ({image_tag})")

            # baking `buildah unshare` command to wrap other buildah commands with
            # so that container does not need to be running in a privileged mode to be able
            # to function
            buildah_unshare_command = sh.buildah.bake('unshare')  # pylint: disable=no-member

            # mount the container filesystem and get mount path
            #
            # NOTE: run in the context of `buildah unshare` so that container does not
            #       need to be run in a privileged mode
            print(f"\nMount container: {container_name}")
            container_mount_path = mount_container(
                buildah_unshare_command=buildah_unshare_command,
                container_id=container_name
            )
            print(f"Mounted container ({container_name}) with mount path: '{container_mount_path}'")

            try:
                # download the open scap input file
                oscap_input_definitions_uri = self.get_value('oscap-input-definitions-uri')
                print(f"\nDownload input definitions: {oscap_input_definitions_uri}")
                oscap_input_file = download_and_decompress_source_to_destination(
                    source_uri=oscap_input_definitions_uri,
                    destination_dir=self.work_dir_path
                )
                print(f"Downloaded input definitions to: {oscap_input_file}")
            except (RuntimeError, ValueError) as error:
                raise StepRunnerException(
                    f"Error downloading OpenSCAP input file: {error}"
                ) from error

            try:
                # if specified download oscap tailoring file
                oscap_tailoring_file = None
                oscap_tailoring_file_uri = self.get_value('oscap-tailoring-uri')
                if oscap_tailoring_file_uri:
                    print(f"\nDownload oscap tailoring file: {oscap_tailoring_file_uri}")
                    oscap_tailoring_file = download_and_decompress_source_to_destination(
                        source_uri=oscap_tailoring_file_uri,
                        destination_dir=self.work_dir_path
                    )
                    print(f"Download oscap tailoring file to: {oscap_tailoring_file}")
            except (RuntimeError, ValueError) as error:
                raise StepRunnerException(
                    f"Error downloading OpenSCAP tailoring file: {error}"
                ) from error

            # determine oscap eval type based on document type
            print(f"\nDetermine OpenSCAP document type of input file: {oscap_input_file}")
            oscap_document_type = OpenSCAPGeneric.__get_oscap_document_type(
                oscap_input_file=oscap_input_file
            )
            print(
                "Determined OpenSCAP document type of input file"
                f" ({oscap_input_file}): {oscap_document_type}"
            )
            print(
                f"\nDetermine OpenSCAP eval type for input file ({oscap_input_file}) "
                f"of document type: {oscap_document_type}"
            )
            oscap_eval_type = OpenSCAPGeneric.__get_oscap_eval_type_based_on_document_type(
                oscap_document_type=oscap_document_type
            )
            print(
                "Determined OpenSCAP eval type of input file"
                f" ({oscap_input_file}): {oscap_eval_type}"
            )

            # Execute scan in the context of buildah unshare
            #
            # NOTE: run in the context of `buildah unshare` so that container does not
            #       need to be run in a privilaged mode
            oscap_out_file_path = self.write_working_file(f'oscap-{oscap_eval_type}-out')
            oscap_xml_results_file_path = self.write_working_file(
                f'oscap-{oscap_eval_type}-results.xml'
            )
            oscap_html_report_path = self.write_working_file(f'oscap-{oscap_eval_type}-report.html')
            print("\nRun oscap scan")
            oscap_eval_success, oscap_eval_fails = OpenSCAPGeneric.__run_oscap_scan(
                buildah_unshare_command=buildah_unshare_command,
                oscap_eval_type=oscap_eval_type,
                oscap_input_file=oscap_input_file,
                oscap_out_file_path=oscap_out_file_path,
                oscap_xml_results_file_path=oscap_xml_results_file_path,
                oscap_html_report_path=oscap_html_report_path,
                container_mount_path=container_mount_path,
                oscap_profile=oscap_profile,
                oscap_tailoring_file=oscap_tailoring_file,
                oscap_fetch_remote_resources=oscap_fetch_remote_resources
            )
            print(f"OpenSCAP scan completed with eval success: {oscap_eval_success}")

            # save scan results
            step_result.success = oscap_eval_success
            if not oscap_eval_success:
                step_result.message = f"OSCAP eval found issues:\n{oscap_eval_fails}"

            step_result.add_artifact(
                name='html-report',
                value=oscap_html_report_path
            )
            step_result.add_artifact(
                name='xml-report',
                value=oscap_xml_results_file_path
            )
            step_result.add_artifact(
                name='stdout-report',
                value=oscap_out_file_path
            )
        except (StepRunnerException, RuntimeError) as error:
            step_result.success = False
            step_result.message = str(error)

        return step_result

    @staticmethod
    def __get_oscap_document_type(oscap_input_file):
        """Gets the OpenSCAP document type for a given input file.

        Parameters
        ----------
        oscap_input_file : path
            Path to OSCAP file to determine the OpenSCAP document type of.

        Returns
        -------
        str
            OpenSCAP document type. For example:
            * Source Data Stream
            * XCCDF Checklist
            * OVAL Definitions

        Raises
        ------
        StepRunnerException
            If error getting document type of oscap input file.
        """

        oscap_document_type = None
        try:
            oscap_info_out_buff = StringIO()
            sh.oscap.info(  # pylint: disable=no-member
                oscap_input_file,
                _out=oscap_info_out_buff
            )
            oscap_info_out = oscap_info_out_buff.getvalue().rstrip()
            oscap_document_type_match = OpenSCAPGeneric.OSCAP_INFO_DOC_TYPE_PATTERN.search(
                oscap_info_out
            )
            oscap_document_type = oscap_document_type_match.groupdict()['doctype']
        except sh.ErrorReturnCode as error:
            raise StepRunnerException(
                f"Error getting document type of oscap input file"
                f" ({oscap_input_file}): {error}"
            ) from error

        return oscap_document_type

    @staticmethod
    def __get_oscap_eval_type_based_on_document_type(oscap_document_type):
        """Given an OSCAP document type returns the type of oscap eval that should be used.

        Parameters
        ----------
        oscap_document_type : str
            OSCAP Document type to get the oscap eval type for.

        Returns
        -------
        str
            OSCAP eval type to perform on document with given oscap document type.
        """
        oscap_eval_type = None

        if oscap_document_type == 'Source Data Stream':
            oscap_eval_type = 'xccdf'
        elif oscap_document_type == 'XCCDF Checklist':
            oscap_eval_type = 'xccdf'
        elif oscap_document_type == 'OVAL Definitions':
            oscap_eval_type = 'oval'

        return oscap_eval_type

    @staticmethod
    def __run_oscap_scan(  # pylint: disable=too-many-arguments,too-many-locals,too-many-branches,too-many-statements
        buildah_unshare_command,
        oscap_eval_type,
        oscap_input_file,
        oscap_out_file_path,
        oscap_xml_results_file_path,
        oscap_html_report_path,
        container_mount_path,
        oscap_profile=None,
        oscap_tailoring_file=None,
        oscap_fetch_remote_resources=True
    ):
        """Run an oscap scan in the context of a buildah unshare to run "rootless".

        Parameters
        ----------
        buildah_unshare_command : sh.buildah.unshare.bake()
            A baked sh.buildah.unshare command to use to run this command in the context off
            so that this can be done "rootless".
        oscap_eval_type : str
            The type of oscap eval to perform. Must be a valid oscap eval type.
            EX: xccdf, oval
        oscap_input_file : str
            Path to rules file passed to the oscap command.
        oscap_out_file_path : str
            Path to write the stdout and stderr of running the oscap command to.
        oscap_xml_results_file_path : str
            Write the scan results into this file.
        oscap_html_report_path : str
            Write the human readable (HTML) report into this file.
        container_mount_path : str
            Path to the mounted container to scan.
        oscap_tailoring_file : str
            XCCF Tailoring file.
            See:
            - https://www.open-scap.org/security-policies/customization/
            - https://www.open-scap.org/resources/documentation/customizing-scap-security-guide-for-your-use-case/ # pylint: disable=line-too-long
            - https://static.open-scap.org/openscap-1.2/oscap_user_manual.html#_how_to_tailor_source_data_stream # pylint: disable=line-too-long
        oscap_profile : str
            OpenSCAP profile to evaluate. Must be a valid profile in the given oscap_input_file.
            EX: if you perform an `oscap info oscap_input_file` the profile must be listed.

        Returns
        -------
        oscap_eval_success : bool
            True if oscap eval passed all rules
            False if oscap eval failed any rules
        oscap_eval_fails : str
            If oscap_eval_success is True then indeterminate.
            If oscap_eval_success is False then string of all of the failed rules.

        Raises
        ------
        StepRunnerException
            If unexpected error running oscap scan.
        """

        oscap_profile_flag = None
        if oscap_profile is not None:
            oscap_profile_flag = f"--profile={oscap_profile}"

        oscap_fetch_remote_resources_flag = None
        if isinstance(oscap_fetch_remote_resources, str):
            oscap_fetch_remote_resources = strtobool(oscap_fetch_remote_resources)
        if oscap_fetch_remote_resources:
            oscap_fetch_remote_resources_flag = "--fetch-remote-resources"

        oscap_tailoring_file_flag = None
        if oscap_tailoring_file is not None:
            oscap_tailoring_file_flag = f"--tailoring-file={oscap_tailoring_file}"

        oscap_eval_success = None
        oscap_eval_out_buff = StringIO()
        oscap_eval_out = ""
        oscap_eval_fails = None
        try:
            oscap_chroot_command = buildah_unshare_command.bake("oscap-chroot")
            with open(oscap_out_file_path, 'w', encoding='utf-8') as oscap_out_file:
                out_callback = create_sh_redirect_to_multiple_streams_fn_callback([
                    oscap_eval_out_buff,
                    oscap_out_file
                ])
                err_callback = create_sh_redirect_to_multiple_streams_fn_callback([
                    oscap_eval_out_buff,
                    oscap_out_file
                ])
                oscap_chroot_command(
                    container_mount_path,
                    oscap_eval_type,
                    'eval',
                    oscap_profile_flag,
                    oscap_fetch_remote_resources_flag,
                    oscap_tailoring_file_flag,
                    f'--results={oscap_xml_results_file_path}',
                    f'--report={oscap_html_report_path}',
                    oscap_input_file,
                    _out=out_callback,
                    _err=err_callback,
                    _tee='err'
                )
                oscap_eval_success = True
        except sh.ErrorReturnCode_1 as error:  # pylint: disable=no-member
            oscap_eval_success = error
        except sh.ErrorReturnCode_2 as error:  # pylint: disable=no-member
            # XCCDF: If there is at least one rule with either fail or unknown result,
            #           oscap-scan finishes with return code 2.
            # OVAL:  Never returned
            #
            # Source: https://www.systutorials.com/docs/linux/man/8-oscap/
            if oscap_eval_type == 'xccdf':
                oscap_eval_success = False
            else:
                oscap_eval_success = error
        except sh.ErrorReturnCode as error:
            oscap_eval_success = error

        # get the oscap output
        oscap_eval_out = oscap_eval_out_buff.getvalue()

        # parse the oscap output
        # NOTE: oscap is puts carrage returns (\r / ^M) in their output, remove them
        oscap_eval_out = re.sub('\r', '', oscap_eval_out)

        # print the oscap output no matter the results
        print(oscap_eval_out)

        # if unexpected error throw error
        if isinstance(oscap_eval_success, Exception):
            raise StepRunnerException(
                f"Error running 'oscap {oscap_eval_type} eval': {oscap_eval_success} "
            ) from oscap_eval_success

        # NOTE: oscap oval eval returns exit code 0 whether or not any rules failed
        #       need to search output to determine if there were any rule failures
        if oscap_eval_type == 'oval' and oscap_eval_success:
            oscap_eval_fails = ""
            for match in OpenSCAPGeneric.OSCAP_OVAL_STDOUT_PATTERN.finditer(oscap_eval_out):
                # NOTE: need to do regex and not == because may contain xterm color chars
                if OpenSCAPGeneric.OSCAP_OVAL_STDOUT_FAIL_PATTERN.search(
                        match.groupdict()['ruleresult']
                ):
                    oscap_eval_fails += match.groupdict()['ruleblock']
                    oscap_eval_fails += "\n"
                    oscap_eval_success = False

        # if failed xccdf eval then parse out the fails
        if oscap_eval_type == 'xccdf' and not oscap_eval_success:
            oscap_eval_fails = ""
            for match in OpenSCAPGeneric.OSCAP_XCCDF_STDOUT_PATTERN.finditer(oscap_eval_out):
                # NOTE: need to do regex and not == because may contain xterm color chars
                if re.search(r'fail', match.groupdict()['ruleresult']):
                    oscap_eval_fails += "\n"
                    oscap_eval_fails += match.groupdict()['ruleblock']
                    oscap_eval_fails += "\n"

        return oscap_eval_success, oscap_eval_fails



import os
from unittest.mock import patch

from ploigos_step_runner import StepResult, StepRunnerException, WorkflowResult
from ploigos_step_runner.step_implementers.unit_test import MavenTest
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase
from tests.helpers.test_utils import Any

class BaseTestStepImplementerMavenTest(
    BaseStepImplementerTestCase
):
    def create_step_implementer(
        self,
        step_config={},
        workflow_result=None,
        parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=MavenTest,
            step_config=step_config,
            step_name='unit-test',
            implementer='MavenTest',
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

@patch("ploigos_step_runner.step_implementers.shared.MavenGeneric.__init__")
class TestStepImplementerMavenTest___init__(BaseStepImplementerTestCase):
    def test_defaults(self, mock_super_init):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        config = {}

        MavenTest(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config
        )

        mock_super_init.assert_called_once_with(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=None,
            maven_phases_and_goals=['test']
        )

    def test_given_environment(self, mock_super_init):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        config = {}

        MavenTest(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment='mock-env'
        )

        mock_super_init.assert_called_once_with(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment='mock-env',
            maven_phases_and_goals=['test']
        )

class TestStepImplementerMavenTest_step_implementer_config_defaults(
    BaseStepImplementerTestCase
):
    def test_result(self):
        self.assertEqual(
            MavenTest.step_implementer_config_defaults(),
            {
                'pom-file': 'pom.xml',
                'tls-verify': True,
                'maven-profiles': [],
                'maven-additional-arguments': [],
                'maven-no-transfer-progress': True
            }
        )

class TestStepImplementerMavenTest__required_config_or_result_keys(
    BaseStepImplementerTestCase
):
    def test_result(self):
        self.assertEqual(
            MavenTest._required_config_or_result_keys(),
            [
                'pom-file'
            ]
        )

@patch.object(MavenTest, '_run_maven_step')
@patch.object(MavenTest, 'write_working_file', return_value='/mock/mvn_output.txt')
@patch.object(MavenTest, '_MavenTest__get_test_report_dir', return_value='/mock/test-results-dir')
@patch.object(MavenTest, '_gather_evidence_from_test_report_directory_testsuite_elements')
class TestStepImplementerMavenTest__run_step(
    BaseTestStepImplementerMavenTest
):
    def create_step_implementer(
        self,
        step_config={},
        workflow_result=None,
        parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=MavenTest,
            step_config=step_config,
            step_name='unit-test',
            implementer='MavenTest',
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

    def test_success_with_report_dir(
        self,
        mock_gather_evidence,
        mock_get_test_report_dir,
        mock_write_working_file,
        mock_run_maven_step
    ):
        with TempDirectory() as test_dir:
            # setup test
            parent_work_dir_path = os.path.join(test_dir.path, 'working')
            pom_file = os.path.join(test_dir.path, 'mock-pom.xml')
            step_config = {
                'pom-file': pom_file
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # run test
            actual_step_result = step_implementer._run_step()

            # verify results
            expected_step_result = StepResult(
                step_name='unit-test',
                sub_step_name='MavenTest',
                sub_step_implementer_name='MavenTest'
            )
            expected_step_result.add_artifact(
                description="Standard out and standard error from maven.",
                name='maven-output',
                value='/mock/mvn_output.txt'
            )
            expected_step_result.add_artifact(
                description="Test report generated when running unit tests.",
                name='test-report',
                value='/mock/test-results-dir'
            )
            self.assertEqual(actual_step_result, expected_step_result)

            mock_run_maven_step.assert_called_once_with(
                mvn_output_file_path='/mock/mvn_output.txt'
            )
            mock_gather_evidence.assert_called_once_with(
                step_result=Any(StepResult),
                test_report_dir='/mock/test-results-dir'
            )

    def test_success_no_report_dir(
        self,
        mock_gather_evidence,
        mock_get_test_report_dir,
        mock_write_working_file,
        mock_run_maven_step
    ):
        with TempDirectory() as test_dir:
            # setup test
            parent_work_dir_path = os.path.join(test_dir.path, 'working')
            pom_file = os.path.join(test_dir.path, 'mock-pom.xml')
            step_config = {
                'pom-file': pom_file
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # setup mocks
            mock_get_test_report_dir.return_value = None

            # run test
            actual_step_result = step_implementer._run_step()

            # verify results
            expected_step_result = StepResult(
                step_name='unit-test',
                sub_step_name='MavenTest',
                sub_step_implementer_name='MavenTest'
            )
            expected_step_result.add_artifact(
                description="Standard out and standard error from maven.",
                name='maven-output',
                value='/mock/mvn_output.txt'
            )
            self.assertEqual(actual_step_result, expected_step_result)

            mock_run_maven_step.assert_called_once_with(
                mvn_output_file_path='/mock/mvn_output.txt'
            )
            mock_gather_evidence.assert_not_called()

    def test_fail_with_report_dir(
        self,
        mock_gather_evidence,
        mock_get_test_report_dir,
        mock_write_working_file,
        mock_run_maven_step
    ):
        with TempDirectory() as test_dir:
            # setup test
            parent_work_dir_path = os.path.join(test_dir.path, 'working')
            pom_file = os.path.join(test_dir.path, 'mock-pom.xml')
            step_config = {
                'pom-file': pom_file
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # setup mocks
            mock_run_maven_step.side_effect = StepRunnerException('mock error')

            # run test
            actual_step_result = step_implementer._run_step()

            # verify results
            expected_step_result = StepResult(
                step_name='unit-test',
                sub_step_name='MavenTest',
                sub_step_implementer_name='MavenTest'
            )
            expected_step_result.success = False
            expected_step_result.message = "Error running maven. " \
                "More details maybe found in report artifacts: " \
                "mock error"
            expected_step_result.add_artifact(
                description="Standard out and standard error from maven.",
                name='maven-output',
                value='/mock/mvn_output.txt'
            )
            expected_step_result.add_artifact(
                description="Test report generated when running unit tests.",
                name='test-report',
                value='/mock/test-results-dir'
            )
            self.assertEqual(actual_step_result, expected_step_result)

            mock_run_maven_step.assert_called_once_with(
                mvn_output_file_path='/mock/mvn_output.txt'
            )
            mock_gather_evidence.assert_called_once_with(
                step_result=Any(StepResult),
                test_report_dir='/mock/test-results-dir'
            )

    def test_fail_no_report_dir(
        self,
        mock_gather_evidence,
        mock_get_test_report_dir,
        mock_write_working_file,
        mock_run_maven_step
    ):
        with TempDirectory() as test_dir:
            # setup test
            parent_work_dir_path = os.path.join(test_dir.path, 'working')
            pom_file = os.path.join(test_dir.path, 'mock-pom.xml')
            step_config = {
                'pom-file': pom_file
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # setup mocks
            mock_run_maven_step.side_effect = StepRunnerException('mock error')
            mock_get_test_report_dir.return_value = None

            # run test
            actual_step_result = step_implementer._run_step()

            # verify results
            expected_step_result = StepResult(
                step_name='unit-test',
                sub_step_name='MavenTest',
                sub_step_implementer_name='MavenTest'
            )
            expected_step_result.success = False
            expected_step_result.message = "Error running maven. " \
                "More details maybe found in report artifacts: " \
                "mock error"
            expected_step_result.add_artifact(
                description="Standard out and standard error from maven.",
                name='maven-output',
                value='/mock/mvn_output.txt'
            )
            self.assertEqual(actual_step_result, expected_step_result)

            mock_run_maven_step.assert_called_once_with(
                mvn_output_file_path='/mock/mvn_output.txt'
            )
            mock_gather_evidence.assert_not_called()

@patch.object(MavenTest, '_attempt_get_test_report_directory')
class TestStepImplementerMavenTest___get_test_report_dir(
    BaseTestStepImplementerMavenTest
):
    def test_user_given_test_reports_dir(self, mock_attempt_get_test_report_directory):
        with TempDirectory() as test_dir:
            # setup test
            parent_work_dir_path = os.path.join(test_dir.path, 'working')
            step_config = {
                'test-reports-dir': '/mock/user-given/test-reports-dir'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # run test
            actual_test_report_dir = step_implementer._MavenTest__get_test_report_dir()

            # verify results
            self.assertEqual(actual_test_report_dir, '/mock/user-given/test-reports-dir')
            mock_attempt_get_test_report_directory.assert_not_called()

    def test_dynamically_determined_test_reports_dir(self, mock_attempt_get_test_report_directory):
        with TempDirectory() as test_dir:
            # setup test
            parent_work_dir_path = os.path.join(test_dir.path, 'working')
            step_config = {}
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # setup mocks
            mock_attempt_get_test_report_directory.return_value = \
                '/mock/dynamically-determined/test-reports-dir'

            # run test
            actual_test_report_dir = step_implementer._MavenTest__get_test_report_dir()

            # verify results
            self.assertEqual(
                actual_test_report_dir,
                '/mock/dynamically-determined/test-reports-dir'
            )
            mock_attempt_get_test_report_directory.assert_called_once_with(
                plugin_name='maven-surefire-plugin',
                configuration_key='reportsDirectory',
                default='target/surefire-reports'
            )

    def test_dynamically_determined_test_reports_dir_errors(self, mock_attempt_get_test_report_directory):
        with TempDirectory() as test_dir:
            # setup test
            parent_work_dir_path = os.path.join(test_dir.path, 'working')
            step_config = {}
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # setup mocks
            mock_attempt_get_test_report_directory.side_effect = StepRunnerException('mock error')

            # run test
            actual_test_report_dir = step_implementer._MavenTest__get_test_report_dir()

            # verify results
            self.assertEqual(actual_test_report_dir, None)
            mock_attempt_get_test_report_directory.assert_called_once_with(
                plugin_name='maven-surefire-plugin',
                configuration_key='reportsDirectory',
                default='target/surefire-reports'
            )



"""Abstract class and helper constants for StepImplementer.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:
* static configuration
* runtime configuration
* previous step results

Configuration Key       | Required? | Default | Description
------------------------|-----------|---------|-----------
`additional-artifacts`  | No        | `[]`    | List of additional artifacts, or list of dicts \
                                                of additional artifacts to add as artifacts on the \
                                                step step result (pass or fail). \
<br/><b>EX 1</b>: <br/>\
<pre>additional-artifacts: <br/>\
- randomValue1 <br/>\
- /path/to/something/important_dir <br/>\
- /path/to/cool/file.xml</pre> \
<br/><b>EX 1</b>: <br/>\
<pre>additional-artifacts: <br/>\
- name: mock-value <br/>\
  value: randomValue1 <br/>\
- name: mock-dir <br/>\
  value: /path/to/something/important_dir <br/>\
- name: mock-file <br/>\
  value: /path/to/cool/file.xml</pre>
"""# pylint: disable=line-too-long
import json
import os
import pprint
import sys
import textwrap
from abc import ABC, abstractmethod
from contextlib import redirect_stderr, redirect_stdout
from pathlib import Path

from ploigos_step_runner.config.config_value import ConfigValue
from ploigos_step_runner import StepResult
from ploigos_step_runner.utils.io import TextIOIndenter


class DefaultSteps:  # pylint: disable=too-few-public-methods
    """Convenience constants for the default steps used in the default workflow definition.
    """
    GENERATE_METADATA = 'generate-metadata'
    TAG_SOURCE = 'tag-source'
    STATIC_CODE_ANALYSIS = 'static-code-analysis'
    PACKAGE = 'package'
    UNIT_TEST = 'unit-test'
    PUSH_ARTIFACTS = 'push-artifacts'
    CREATE_CONTAINER_IMAGE = 'create-container-image'
    PUSH_CONTAINER_IMAGE = 'push-container-image'
    SIGN_CONTAINER_IMAGE = 'sign-container-image'
    CONTAINER_IMAGE_UNIT_TEST = 'container-image-unit-test'
    CONTAINER_IMAGE_STATIC_COMPLIANCE_SCAN = 'container-image-static-compliance-scan'
    CONTAINER_IMAGE_STATIC_VULNERABILITY_SCAN = 'container-image-static-vulnerability-scan'
    CREATE_DEPLOYMENT_ENVIRONMENT = 'create-deployment-environment'
    DEPLOY = 'deploy'
    UAT = 'uat'
    RUNTIME_VULNERABILITY_SCAN = 'runtime-vulnerability-scan'
    CANARY_TEST = 'canary-test'
    PUBLISH_WORKFLOW_RESULTS = 'publish-workflow-results'


class StepImplementer(ABC):  # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-public-methods
    """Abstract representation of a step implementer.

    Parameters
    ----------
    workflow_result : WorkflowResult
        The WorkflowResult to add StepResult to when running this StepImplementer.
    parent_work_dir_path : str
        Path to the directory to write step working files to
    step_environment_config : dict, optional
        Step configuration specific to the current environment.
    config : SubStepConfig
        Configuration for this step.
    environment : str
        Environment name to execute this step against

    Attributes
    __config : SubStepConfig
    __environment : str
    """

    __TITLE_LENGTH = 80
    __INDENT_SIZE   = 4

    def __init__(  # pylint: disable=too-many-arguments
        self,
        workflow_result,
        parent_work_dir_path,
        config,
        environment=None
    ):
        self.__parent_work_dir_path = parent_work_dir_path

        self.__config = config
        self.__environment = environment

        self.__workflow_result = workflow_result

        super().__init__()

    @property
    def config(self):
        """
        Returns
        -------
        SubStepConfig
            Configuration for this step.
        """
        return self.__config

    @property
    def environment(self):
        """
        Returns
        -------
        str
            Environment name to execute this step against
        """
        return self.__environment

    @property
    def step_config(self):
        """
        Returns
        -------
        dict
            Step configuration.
        """
        return self.config.sub_step_config

    @property
    def step_config_overrides(self):
        """
        Returns
        -------
        dict
            Step configuration overrides.
        """
        return self.config.step_config_overrides

    @property
    def step_environment_config(self):
        """
        Returns
        -------
        dict
            Step configuration specific to the current environment.
        """
        return self.config.get_sub_step_env_config(self.environment)

    @property
    def global_config_defaults(self):
        """
        Returns
        -------
        dict
            Global configuration defaults affecting this step.
        """
        return self.config.global_defaults

    @property
    def global_environment_config_defaults(self):
        """
        Returns
        -------
        dict
            Global configuration defaults affecting this step specific to the current environment.
        """

        if self.environment is not None:
            defaults_for_env = self.config.get_global_environment_defaults(self.environment)
        else:
            defaults_for_env = {}

        return defaults_for_env

    @property
    def work_dir_path(self):
        """Get the OS path to the working folder for this step.
        If the folder does not exist, create it.

        Returns
        -------
        str
            OS path to the working folder plus step name.
        """
        work_dir_path_step = os.path.join(self.__parent_work_dir_path, self.step_name)
        if self.environment:
            work_dir_path_step = os.path.join(work_dir_path_step, self.environment)
        os.makedirs(work_dir_path_step, exist_ok=True)
        return work_dir_path_step


    @property
    def step_name(self):
        """
        Getter for the step name implemented by this step.
        Returns
        -------
        str
            Step name implemented by this step.
        """
        return self.config.step_name

    @property
    def sub_step_name(self):
        """
        Getter for the sub step name implemented by this step.
        Returns
        -------
        str
            Sub step name.
        """
        return self.config.sub_step_name

    @property
    def sub_step_implementer_name(self):
        """
        Getter for the sub step implementer name implemented by this step.
        Returns
        -------
        str
            Sub step implementer name.
        """
        return self.config.sub_step_implementer_name

    @property
    def workflow_result(self):
        """
        Returns
        -------
        WorkflowResult
            Object containing a list of dictionary of step results
            from previous steps.
        """
        return self.__workflow_result

    @staticmethod
    @abstractmethod
    def step_implementer_config_defaults():
        """
        Getter for the StepImplementer's configuration defaults.

        Notes
        -----
        These are the lowest precedence configuration values.

        Returns
        -------
        dict
            Default values to use for step configuration values.
        """

    @staticmethod
    @abstractmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """

    @abstractmethod
    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """

    def _validate_required_config_or_previous_step_result_artifact_keys(self):
        """Validates that the required configuration keys or previous step result artifacts
        are set and have valid values.

        Raises
        ------
        AssertionError
            If step configuration or previous step result artifacts have invalid required values.
        """
        invalid_required_keys = []
        for required_key in self._required_config_or_result_keys():
            required_value = self.get_value(required_key)

            if required_value is None:
                invalid_required_keys.append(required_key)

        assert (not invalid_required_keys), \
            'Missing required step configuration or previous step result artifact keys: ' + \
            f'{invalid_required_keys}'

    def run_step(self):
        """Wrapper for running the implemented step.

        Returns
        -------
        StepResult
            Results of running this step.
        """

        StepImplementer.__print_section_title(
            f"Step Start - {self.step_name} ({self.sub_step_name})"
        )

        # print information about the configuration
        StepImplementer.__print_section_title(
            f"Configuration - {self.step_name} ({self.sub_step_name})",
            div_char="-",
            indent=1
        )
        StepImplementer.__print_data(
            "Step Implementer Configuration Defaults",
            ConfigValue.convert_leaves_to_values(self.step_implementer_config_defaults())
        )
        StepImplementer.__print_data(
            "Global Configuration Defaults",
            ConfigValue.convert_leaves_to_values(self.global_config_defaults)
        )
        StepImplementer.__print_data(
            "Global Environment Configuration Defaults",
            ConfigValue.convert_leaves_to_values(self.global_environment_config_defaults)
        )
        StepImplementer.__print_data(
            "Step Configuration",
            ConfigValue.convert_leaves_to_values(self.step_config)
        )
        StepImplementer.__print_data(
            "Step Environment Configuration",
            ConfigValue.convert_leaves_to_values(self.step_environment_config)
        )
        StepImplementer.__print_data(
            "Step Configuration Runtime Overrides",
            ConfigValue.convert_leaves_to_values(self.step_config_overrides)
        )

        # create the munged runtime step configuration and print
        copy_of_runtime_step_config = self.get_copy_of_runtime_step_config()
        StepImplementer.__print_data(
            "Runtime Step Configuration",
            ConfigValue.convert_leaves_to_values(copy_of_runtime_step_config)
        )

        step_result = None
        try:
            # validate the runtime step configuration
            self._validate_required_config_or_previous_step_result_artifact_keys()

            # run the step
            StepImplementer.__print_section_title(
                f"Standard Out - {self.step_name} ({self.sub_step_name})",
                div_char="-",
                indent=1
            )
            indented_stdout = TextIOIndenter(parent_stream=sys.stdout, indent_level=2)
            indented_stderr = TextIOIndenter(parent_stream=sys.stderr, indent_level=2)
            with redirect_stdout(indented_stdout), redirect_stderr(indented_stderr):
                step_result = self._run_step()
                sys.stdout.flush()
                sys.stderr.flush()

            # add any additional artifacts
            self.__add_additional_artifacts_to_step_result(step_result=step_result)
        except AssertionError as invalid_error:
            step_result = StepResult.from_step_implementer(self)
            step_result.success = False
            step_result.message = str(invalid_error)

        # print the step run results
        StepImplementer.__print_section_title(
            f"Results - {self.step_name} ({self.sub_step_name})",
            div_char="-",
            indent=1
        )
        StepImplementer.__print_data('Environment', step_result.environment)
        StepImplementer.__print_data('Step', step_result.step_name)
        StepImplementer.__print_data('Sub Step', step_result.sub_step_name)
        StepImplementer.__print_data('Sub Step Implementer', step_result.sub_step_implementer_name)
        StepImplementer.__print_data('Success', step_result.success)
        StepImplementer.__print_data('Message', step_result.message)
        StepImplementer.__print_data('Artifacts', step_result.artifacts_dicts)
        StepImplementer.__print_data('Evidence', step_result.evidence_dicts)

        StepImplementer.__print_section_title(f'Step End - {self.step_name} ({self.sub_step_name})')
        return step_result

    def get_value(self, key):
        """Get the value for a given key, either from given configuration or from the result
        of any previous step.

        Parameters
        ----------
        key : str, array
            Key to get the configuration value or result value for.
            If array will get configuration value or result value for first key in list with a
            value.

        Returns
        -------
        str, int, dict, list, or bool or None
            Configuration value, or if not set, result value from any previous step for the given
            key. Or None if not found.
        """

        if isinstance(key, list):
            keys = key
        else:
            keys = [key]

        for k in keys:
            # first try to get config value
            config_value = self.get_config_value(k)
            if config_value is not None:
                return config_value

            # if not found config value try to get result value specific to current environment
            if self.environment:
                result_value = self.get_result_value(
                    artifact_name=k,
                    environment=self.environment
                )
                if result_value is not None:
                    return result_value

            # last try getting result value from no specific environment
            result_value = self.get_result_value(
                artifact_name=k
            )

            if result_value:
                break

        return result_value

    def get_config_value(self, key):
        """Convenience function for self.config.get_config_value.

        Get the configuration value for a given configuration key from the
        merged set of configuration sources.

        From least precedence to highest precedence.

            1. defaults
            2. Global Configuration Defaults (self.global_config_defaults)
            3. Global Environment Configuration Defaults (self.global_environment_config_defaults)
            4. Step Configuration ( self.step_config)
            5. Step Environment Configuration (self.step_environment_config)
            6. Step Configuration Runtime Overrides (step_config_runtime_overrides)

        Also See
        --------
        SubStepConfig.get_config_value
        get_copy_of_runtime_step_config

        Parameters
        ----------
        key : str
            Key to get the configuration value for.

        Returns
        -------
        str, int, dict, list, or bool or None
            Value of the given configuration key or None if one does not exist
            for this sub step in the given context with the given defaults.
        """
        return self.config.get_config_value(
            key,
            self.environment,
            self.step_implementer_config_defaults())

    def get_copy_of_runtime_step_config(self):
        """Convenience function for self.config.get_copy_of_runtime_step_config

        See Also
        --------
        SubStepConfig.get_copy_of_runtime_step_config
        get_config_value

        Returns
        -------
        dict
            A deep copy of the merged runtime step configuration
        """
        return self.config.get_copy_of_runtime_step_config(
            self.environment,
            self.step_implementer_config_defaults())

    def has_config_value(self, keys, match_any=False):
        """Determines if step has values for any of the given keys.

        Parameters
        ----------
        keys : str or list of str
            Keys to check to see if there are values for any or all of these key(s).
        match_any : bool
            If True only one key has to have a value.
            If False then all given keys have to have values.

        Returns
        -------
        bool
            True if any/all (depending on value of match_any) keys have a value.
            False otherwise.
        """
        if isinstance(keys, str):
            keys = [keys]

        result = not match_any
        for key in keys:
            if match_any and self.get_value(key) is not None:
                result = True
                break

            if not match_any and self.get_value(key) is None:
                result = False
                break

        return result

    def get_result_value(
        self,
        artifact_name,
        step_name=None,
        sub_step_name=None,
        environment=None
    ):
        """Get the value for the named artifact from a previous step result.

        If step_name is provided,
            search for artifact_name in step_name only
        If step_name and sub_step_name is provided,
            search for artifact_name in step_name/sub_step only
        Otherwise,
             search for the FIRST occurrence of the artifact_name

        EG:  Artifacts are a dictionary; return the contents of the value key.
          In this example, 'foo' is the artifact name, return 'bar'
          'foo' : {'description': '', 'value': 'bar'}

        Parameters
        ----------
        artifact_name : str
           Key name of the dictionary.

        Returns
        -------
        str
           Contents of the value for the specified result artifact_name.
        """
        return (
            self.workflow_result.get_artifact_value(
                artifact=artifact_name,
                step_name=step_name,
                sub_step_name=sub_step_name,
                environment=environment
            )
        )

    def create_working_dir_sub_dir(self, sub_dir_relative_path=""):
        """Create a folder under the working/stepname folder.

        Returns
        -------
        str
            Path to created working sub directory.
        """
        file_path = os.path.join(self.work_dir_path, sub_dir_relative_path)
        os.makedirs(file_path, exist_ok=True)
        return file_path

    def write_working_file(self, filename, contents=None):
        """Write content or touch filename in working directory for this step.

        Parameters
        ----------
        filename : str
            File name to create
        contents : str, optional
            Contents to write to the file

        Returns
        -------
        str
            Return a string to the file path
        """
        # eg: step-runner-working/step_name
        file_path = os.path.join(self.work_dir_path, filename)

        # sub-directories might be passed filename, eg: foo/filename
        os.makedirs(os.path.dirname(file_path), exist_ok=True)

        if contents is None:
            Path(file_path).touch()
        else:
            with open(file_path, 'wb') as file:
                file.write(contents)
        return file_path

    def __add_additional_artifacts_to_step_result(self, step_result):
        """Adds user supplied additional artifacts to step result if given.

        Parameters
        ----------
        step_result : StepResult
            StepResult to add the additional artifacts to.
        """
        additional_artifacts = self.get_value('additional-artifacts')
        if isinstance(additional_artifacts, list):
            for additional_artifact in additional_artifacts:
                if isinstance(additional_artifact, dict):
                    step_result.add_artifact(
                        name=additional_artifact['name'],
                        value=additional_artifact['value']
                    )
                else:
                    step_result.add_artifact(
                        name=os.path.basename(additional_artifact),
                        value=additional_artifact
                    )

    @staticmethod
    def __print_section_title(title, div_char="=", indent=0):
        """Utility function for pretty printing section title.

        Parameters
        ----------
        title : str
            Section title to print
        """
        print()
        print()
        StepImplementer.__print_indented(
            text=div_char * StepImplementer.__TITLE_LENGTH,
            indent=indent
        )
        StepImplementer.__print_indented(
            text=title.center(StepImplementer.__TITLE_LENGTH),
            indent=indent
        )
        StepImplementer.__print_indented(
            text=div_char * StepImplementer.__TITLE_LENGTH,
            indent=indent
        )

    @staticmethod
    def __print_data(title, data, indent=2):
        """Utility function for pretty printing data.

        Notes
        -----
        Indent levels are each are StepImplementer.__INDENT_SIZE spaces wide.

        Parameters
        ----------
        title : str
            Title of the data to print.
        data
            Data to print
        indent : int
            Amount to indent the title by and then the content by this +1
        """

        StepImplementer.__print_indented(
            text=title,
            indent=indent
        )

        # NOTE: do a json.dumps so that when printing dicts and arrays the {} and [] get put on
        #       their own lines which there is no way to do with PrettyPrinter.
        if isinstance(data, (dict, list)):
            formated_data = json.dumps(
                data,
                indent=StepImplementer.__INDENT_SIZE
            )
        elif isinstance(data, str):
            formated_data = data
        else:
            printer = pprint.PrettyPrinter()
            formated_data = printer.pformat(data)

        StepImplementer.__print_indented(
            text=formated_data,
            indent=indent + 1
        )
        print()

    @staticmethod
    def __print_indented(text, indent=0):
        """Prints the given text indented by a given indent level.

        Notes
        -----
        Indent levels are each are StepImplementer.__INDENT_SIZE spaces wide.

        Parameters
        ----------
        text : str
            Text to print indented.
        indent : indent
            amount to indent the given text by before printing.
        """
        print(textwrap.indent(
            text=text,
            prefix=" " * (StepImplementer.__INDENT_SIZE * indent)
        ))



import os

from ploigos_step_runner import StepResult, WorkflowResult
from ploigos_step_runner.config.config import Config
from ploigos_step_runner.exceptions import StepRunnerException

from .base_test_case import BaseTestCase


class BaseStepImplementerTestCase(BaseTestCase):
    def create_given_step_implementer(
        self,
        step_implementer,
        step_config={},
        step_name='',
        environment=None,
        implementer='',
        workflow_result=None,
        parent_work_dir_path='',
    ):
        config = Config({
            Config.CONFIG_KEY: {
                step_name: [
                    {
                        'implementer': implementer,
                        'config': step_config
                    }
                ]

            }
        })
        step_config = config.get_step_config(step_name)
        sub_step_config = step_config.get_sub_step(implementer)

        if not workflow_result:
            workflow_result = WorkflowResult()

        step_implementer = step_implementer(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=sub_step_config,
            environment=environment
        )

        return step_implementer

    def setup_previous_result(
        self,
        work_dir_path,
        artifact_config={}
    ):
        step_result = StepResult(
            step_name='test-step',
            sub_step_name='test-sub-step-name',
            sub_step_implementer_name='test-step-implementer-name'
        )
        for key1, val1 in artifact_config.items():
            description = ''
            value = ''
            for key2, val2 in val1.items():
                if key2 == 'description':
                    description = val2
                elif key2 == 'value':
                    value = val2
                else:
                    raise StepRunnerException(f'Given field is not apart of an artifact: {key2}')
            step_result.add_artifact(
                name=key1,
                value=value,
                description=description,
            )
        workflow_result = WorkflowResult()
        workflow_result.add_step_result(step_result=step_result)
        pickle_filename = os.path.join(work_dir_path, 'step-runner-results.pkl')
        workflow_result.write_to_pickle_file(pickle_filename=pickle_filename)

        return workflow_result



"""`StepImplementer` for the `package` step using Maven.

Notes
-----

.. Important::

    If package not specified in pom will default to jar in result.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:

  * static configuration
  * runtime configuration
  * previous step results

Configuration Key            | Required? | Default | Description
-----------------------------|-----------|---------|-----------
`pom-file`                   | Yes       | `'pom.xml'` | pom used when executing maven.
`tls-verify`                 | No        | `True`  | Disables TLS Verification if set to False
`maven-profiles`             | No        | `[]`    | List of maven profiles to use.
`maven-no-transfer-progress` | No        | `True`  | \
                            `True` to suppress the transfer progress of packages maven downloads.
                            `False` to have the transfer progress printed.\
                            See https://maven.apache.org/ref/current/maven-embedder/cli.html
`maven-additional-arguments` | No        | `['-Dmaven.test.skip=true']` \
                                                   | List of additional arguments to use. \
                                                     Skipping tests by default because assuming \
                                                     a previous step already ran them.
`maven-servers`              | No        |         | Dictionary of dictionaries of \
                                                     id, username, password
`maven-repositories`         | No        |         | Dictionary of dictionaries of \
                                                     id, url, snapshots, releases
`maven-mirrors`              | No        |         | Dictionary of dictionaries of \
                                                     id, url, mirror_of
`artifact-extensions`        | Yes       | `["jar", "war", "ear"]` \
                                            | Extensions to look for in the `artifact-parent-dir` \
                                              for built artifacts.
`artifact-parent-dir`        | Yes       | `'target'` \
                                            | Parent directory to look for built artifacts in \
                                              ending in `artifact-extensions`.

Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key | Description
--------------------|------------
`maven-output`      | Path to Stdout and Stderr from invoking Maven.
`packages`          | An array of dictionaries with information on the built artifacts.


## package-artifacts
Keys in the dictionary elements in the `package-artifacts` array in the step results.

| Key             | Description
|-----------------|------------
| `path`          | Absolute path to the built artifact.

"""
import os

from ploigos_step_runner import StepResult, StepRunnerException
from ploigos_step_runner.step_implementers.shared.maven_generic import \
    MavenGeneric

DEFAULT_CONFIG = {
    'maven-additional-arguments': [
        '-Dmaven.test.skip=true'
    ],
    'artifact-extensions': ["jar", "war", "ear"],
    'artifact-parent-dir': 'target'
}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'pom-file'
]

class MavenPackage(MavenGeneric):
    """`StepImplementer` for the `package` step using Maven.
    """
    def __init__(  # pylint: disable=too-many-arguments
        self,
        workflow_result,
        parent_work_dir_path,
        config,
        environment=None
    ):
        super().__init__(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=environment,
            maven_phases_and_goals=['package']
        )

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Returns
        -------
        dict
            Default values to use for step configuration values.

        Notes
        -----
        These are the lowest precedence configuration values.
        """
        return {**MavenGeneric.step_implementer_config_defaults(), **DEFAULT_CONFIG}

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _run_step(self): # pylint: disable=too-many-locals
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        pom_file = self.get_value('pom-file')
        artifact_extensions = self.get_value('artifact-extensions')
        artifact_parent_dir = self.get_value('artifact-parent-dir')

        # package the artifacts
        mvn_output_file_path = self.write_working_file('mvn_output.txt')
        try:
            # execute maven step (params come from config)
            self._run_maven_step(
                mvn_output_file_path=mvn_output_file_path
            )

             # find the artifacts
            packages = []
            pom_file_dir_name = os.path.dirname(os.path.abspath(pom_file))
            files_in_artifact_parent_dir = sorted(os.listdir(os.path.join(
                pom_file_dir_name,
                artifact_parent_dir)))
            for filename in files_in_artifact_parent_dir:
                if any(filename.endswith(ext) for ext in artifact_extensions):
                    packages += [{
                        'path': os.path.join(
                            pom_file_dir_name,
                            artifact_parent_dir,
                            filename
                        )
                    }]

            step_result.add_artifact(
                name='packages',
                value=packages
            )
        except FileNotFoundError as error:
            step_result.success = False
            step_result.message = f"Error finding artifacts after running maven package: {error}"
        except StepRunnerException as error:
            step_result.success = False
            step_result.message = "Error running 'maven package' to package artifacts. " \
                f"More details maybe found in 'maven-output' report artifact: {error}"
        finally:
            step_result.add_artifact(
                description="Standard out and standard error from maven.",
                name='maven-output',
                value=mvn_output_file_path
            )

        return step_result



# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
import os

from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase
from ploigos_step_runner.step_implementers.validate_environment_configuration import \
    ConfiglintFromArgocd
from ploigos_step_runner import StepResult


class TestStepImplementerConfiglintFromArgocd(BaseStepImplementerTestCase):
    def create_step_implementer(
            self,
            step_config={},
            step_name='',
            implementer='',
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=ConfiglintFromArgocd,
            step_config=step_config,
            step_name=step_name,
            implementer=implementer,
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

    # TESTS FOR configuration checks
    def test_step_implementer_config_defaults(self):
        defaults = ConfiglintFromArgocd.step_implementer_config_defaults()
        expected_defaults = {
        }
        self.assertEqual(expected_defaults, defaults)

    def test__required_config_or_result_keys(self):
        required_keys = ConfiglintFromArgocd._required_config_or_result_keys()
        expected_required_keys = [
            'argocd-deployed-manifest'
        ]
        self.assertEqual(required_keys, expected_required_keys)

    def test_run_step_pass(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            test_file_name = 'yamlnotused'
            test_file_path = os.path.join(temp_dir.path, test_file_name)
            temp_dir.write(test_file_path, b'ignored')

            step_config = {}
            artifact_config = {
                'argocd-deployed-manifest': {
                    'value': test_file_path
                }
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='validate-environment-configuration',
                implementer='ConfiglintArgocd',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='validate-environment-configuration',
                sub_step_name='ConfiglintArgocd',
                sub_step_implementer_name='ConfiglintArgocd'
            )

            expected_step_result.add_artifact(
                name='configlint-yml-path',
                value=test_file_path
            )
            self.assertEqual(expected_step_result, result)

    def test_run_step_fail_missing_path_file_from_deploy(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            test_file_name = 'yamlnotused'
            test_file_path = os.path.join(temp_dir.path, test_file_name)
            temp_dir.write(test_file_path, b'ignored')

            step_config = {}
            artifact_config = {
                'argocd-deployed-manifest': {
                    'value': f'{test_file_path}.bad'
                }
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='validate-environment-configuration',
                implementer='ConfiglintArgocd',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='validate-environment-configuration',
                sub_step_name='ConfiglintArgocd',
                sub_step_implementer_name='ConfiglintArgocd'
            )
            expected_step_result.success = False
            expected_step_result.message = f'File specified in argocd-deployed-manifest {test_file_path}.bad not found'
            self.assertEqual(expected_step_result, result)



import unittest
import sh
import os

from .base_test_case import BaseTestCase

class SOPSIntegrationTestCase(BaseTestCase):
    TESTS_GPG_KEY_FINGERPRINT = '6F70E1656E932EFEE8AD898A98871DAE82786C09'

    def setUp(self):
        super().setUp()

        sops_path = sh.which('sops')
        gpg_path = sh.which('gpg')
        if (sops_path is not None) and (gpg_path is not None):
            self.install_gpg_key()
        else:
            self.skipTest('sops and/or pgp not installed')

    def tearDown(self):
        super().tearDown()

        sops_path = sh.which('sops')
        gpg_path = sh.which('gpg')
        if (sops_path is not None) and (gpg_path is not None):
            self.delete_gpg_key()
        else:
            self.skipTest('sops and/or pgp not installed')

    def install_gpg_key(self):
        # install private key
        gpg_private_key_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'ploigos-step-runner-tests-private.asc'
        )
        sh.gpg( # pylint: disable=no-member
            '--import',
            gpg_private_key_path
        )

    def delete_gpg_key(self):
        try:
            # uninstall private key
            sh.gpg( # pylint: disable=no-member
                '--batch',
                '--pinentry-mode',
                'loopback',
                '--yes',
                '--delete-secret-keys',
                SOPSIntegrationTestCase.TESTS_GPG_KEY_FINGERPRINT
            )
        except:
            # don't care if this fails really
            # could fail cuz the test uninstaleld it already
            pass




import os
from pathlib import Path
# from shutil import copyfile
from unittest.mock import PropertyMock, patch
from ploigos_step_runner import StepResult, StepRunnerException, WorkflowResult
from ploigos_step_runner.config.config import Config
from ploigos_step_runner.step_implementers.shared.npm_generic import NpmGeneric
# from ploigos_step_runner.utils.file import create_parent_dir
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase


@patch("ploigos_step_runner.StepImplementer.__init__")
class TestStepImplementerSharedNpmGeneric___init__(BaseStepImplementerTestCase):

    def test_no_environment_no_npm_args(self, mock_super_init):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        config = {}

        step_implementer = NpmGeneric(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config
        )

        self.assertIsNone(step_implementer._NpmGeneric__npm_args)
        mock_super_init.assert_called_once_with(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=None
        )

    def test_with_environment_no_npm_args(self, mock_super_init):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        config = {}

        step_implementer = NpmGeneric(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment='test-env'
        )

        self.assertIsNone(step_implementer._NpmGeneric__npm_args)
        mock_super_init.assert_called_once_with(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment='test-env'
        )

    def test_no_environment_with_npm_args(self, mock_super_init):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        config = {}

        step_implementer = NpmGeneric(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            npm_args=['fake-arg']
        )

        self.assertEqual(
            step_implementer._NpmGeneric__npm_args,
            ['fake-arg']
        )
        mock_super_init.assert_called_once_with(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=None
        )

class TestStepImplementerSharedNpmGeneric_step_implementer_config_defaults(
    BaseStepImplementerTestCase
):
    def test_result(self):
        self.assertEqual(
            NpmGeneric.step_implementer_config_defaults(),
            {
                'package-file': 'package.json',
            }
        )

class TestStepImplementerSharedNpmGeneric__required_config_or_result_keys(
    BaseStepImplementerTestCase
):
    def test_result(self):
        self.assertEqual(
            NpmGeneric._required_config_or_result_keys(),
            [
                'package-file',
                'npm-args'
            ]
        )

class BaseTestStepImplementerSharedNpmGeneric(BaseStepImplementerTestCase):
    def create_step_implementer(
            self,
            step_config={},
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=NpmGeneric,
            step_config=step_config,
            step_name='foo',
            implementer='NpmGeneric',
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

@patch("ploigos_step_runner.StepImplementer._validate_required_config_or_previous_step_result_artifact_keys")
class TestStepImplementerSharedNpmGeneric__validate_required_config_or_previous_step_result_artifact_keys(
    BaseTestStepImplementerSharedNpmGeneric
):
    def test_valid(self, mock_super_validate):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            package_file_path = os.path.join(test_dir.path, 'package.json')
            step_config = {
                'package-file': package_file_path,
                'npm-run-scripts': ['fake-script']
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            Path(package_file_path).touch()
            step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

            mock_super_validate.assert_called_once_with()

    def test_package_file_does_not_exist(self, mock_super_validate):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            package_file_path = '/does/not/exist/package.json'
            step_config = {
                'package-file': package_file_path,
                'npm-run-scripts': 'fake-arg'
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            with self.assertRaisesRegex(
                AssertionError,
                rf'Given npm package file \(package-file\) does not exist: {package_file_path}'
            ):
                step_implementer._validate_required_config_or_previous_step_result_artifact_keys()
                mock_super_validate.assert_called_once_with()

class TestStepImplementerSharedNpmGeneric_config(
    BaseTestStepImplementerSharedNpmGeneric
):
    def test_use_object_property_no_config_value(self):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        config = None

        step_implementer = NpmGeneric(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            npm_args=['run fake:script']
        )

        self.assertEqual(
            step_implementer.npm_args,
            ['run fake:script']
        )

    def test_use_object_property_with_config_value(self):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        step_config = {
            'npm-args': ['config-value-fake-arg']
        }
        config = Config({
            Config.CONFIG_KEY: {
                'foo': [
                    {
                        'implementer': 'NpmGeneric',
                        'config': step_config
                    }
                ]

            }
        })

        step_implementer = NpmGeneric(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            npm_args=['object-property-fake-arg']
        )

        self.assertEqual(
            step_implementer.npm_args,
            ['object-property-fake-arg']
        )

    def test_use_configured_npm_args(self):
        parent_work_dir_path = '/fake/path'
        step_config = {
            'npm-args': ['config-value-npm-args']
        }

        step_implementer = self.create_step_implementer(
            step_config=step_config,
            parent_work_dir_path=parent_work_dir_path,
        )

        self.assertEqual(
            step_implementer.npm_args,
            ['config-value-npm-args']
        )

@patch('ploigos_step_runner.step_implementers.shared.npm_generic.run_npm')
@patch.object(
    NpmGeneric,
    'npm_args',
    new_callable=PropertyMock,
    return_value=['fake-arg']
)

class TestStepImplementerSharedNpmGeneric__run_npm_step(BaseTestStepImplementerSharedNpmGeneric):
    def test_defaults(self, mock_args, mock_run_npm):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            package_file_path = os.path.join(test_dir.path, 'package.json')
            step_config = {
                'package-file': package_file_path
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            npm_output_file_path = os.path.join(test_dir.path, 'npm-output.txt')
            step_implementer._run_npm_step(
                npm_output_file_path=npm_output_file_path
            )

            mock_run_npm.assert_called_with(
                npm_output_file_path=npm_output_file_path,
                npm_args=['fake-arg']
                )

@patch.object(NpmGeneric, '_run_npm_step')
@patch.object(NpmGeneric, 'write_working_file', return_value='/mock/npm_output.txt')
class TestStepImplementerSharedNpmGeneric__run_step(
    BaseTestStepImplementerSharedNpmGeneric):
    def test_success(self, mock_write_working_file, mock_run_npm_step):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            step_config = {}
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # run step
            actual_step_result = step_implementer._run_step()

            # create expected step result
            expected_step_result = StepResult(
                step_name='foo',
                sub_step_name='NpmGeneric',
                sub_step_implementer_name='NpmGeneric'
            )
            expected_step_result.add_artifact(
                description="Standard out and standard error from npm.",
                name='npm-output',
                value='/mock/npm_output.txt'
            )

            # verify step result
            self.assertEqual(
                actual_step_result,
                expected_step_result
            )

            mock_run_npm_step.assert_called_with(
                npm_output_file_path='/mock/npm_output.txt'
            )

    def test_fail(self, mock_write_working_file, mock_run_npm_step):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            step_config = {}
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # run step with mock failure
            mock_run_npm_step.side_effect = StepRunnerException('Mock error running npm')
            actual_step_result = step_implementer._run_step()

            # create expected step result
            expected_step_result = StepResult(
                step_name='foo',
                sub_step_name='NpmGeneric',
                sub_step_implementer_name='NpmGeneric'
            )
            expected_step_result.add_artifact(
                description="Standard out and standard error from npm.",
                name='npm-output',
                value='/mock/npm_output.txt'
            )
            expected_step_result.message = "Error running npm. " \
                "More details maybe found in 'npm-output' report artifact: "\
                "Mock error running npm"
            expected_step_result.success = False

            # verify step result
            self.assertEqual(
                actual_step_result,
                expected_step_result
            )

            mock_write_working_file.assert_called_once()
            mock_run_npm_step.assert_called_with(
                npm_output_file_path='/mock/npm_output.txt'
            )



# pylint: disable=W0614,W0401

"""
Ploigos step runner entry point.

Exit Codes
----------
101
    specified -c/--config must exist and not be empty
102
    specified -c/--config is invalid configuration
200
    step completed with unsuccessful results
300
    step failed completion because of an exception
"""

import argparse
import os.path
import sys
import traceback
from contextlib import redirect_stderr, redirect_stdout

from ploigos_step_runner.config.config import Config
from ploigos_step_runner.decryption_utils import DecryptionUtils
from ploigos_step_runner.step_runner import StepRunner
from ploigos_step_runner.utils.io import TextIOSelectiveObfuscator


def print_error(msg):
    """
    Prints message to STDERR.

    Parameters
    ----------
    msg : string
        Message to print as an error.
    """
    print(msg, file=sys.stderr)


class ParseKeyValueArge(argparse.Action):  # pylint: disable=too-few-public-methods
    """
    https://gist.github.com/fralau/061a4f6c13251367ef1d9a9a99fb3e8d
    """

    def __call__(self, parser, namespace, values, option_string=None):
        key_value_dict = {}

        if values:
            for item in values:
                split_items = item.split("=", 1)
                key = split_items[
                    0
                ].strip()  # we remove blanks around keys, as is logical
                value = split_items[1]

                key_value_dict[key] = value

        setattr(namespace, self.dest, key_value_dict)


def main(argv=None):
    """Main entry point for Ploigos step runner.
    """
    parser = argparse.ArgumentParser(description='Ploigos Step Runner (psr)')
    parser.add_argument(
        '-s',
        '--step',
        required=True,
        help='Workflow step to run'
    )
    parser.add_argument(
        '-e',
        '--environment',
        required=False,
        help='The environment to run this step against.'
    )
    parser.add_argument(
        '-c',
        '--config',
        required=True,
        nargs='+',
        help='Workflow configuration files, or directories containing files, in yml or json'
    )
    parser.add_argument(
        '--step-config',
        metavar='STEP_CONFIG_KEY=STEP_CONFIG_VALUE',
        nargs='+',
        help='Override step config provided by the given config-file with these arguments.',
        action=ParseKeyValueArge
    )
    args = parser.parse_args(argv)

    obfuscated_stdout = TextIOSelectiveObfuscator(sys.stdout)
    obfuscated_stderr = TextIOSelectiveObfuscator(sys.stderr)
    DecryptionUtils.register_obfuscation_stream(obfuscated_stdout)
    DecryptionUtils.register_obfuscation_stream(obfuscated_stderr)

    with redirect_stdout(obfuscated_stdout), redirect_stderr(obfuscated_stderr):
        # validate args
        for config_file in args.config:
            if not os.path.exists(config_file) or os.stat(config_file).st_size == 0:
                print_error('specified -c/--config must exist and not be empty')
                sys.exit(101)

        try:
            config = Config(args.config)
        except (ValueError, AssertionError) as error:
            print_error(f"specified -c/--config is invalid configuration: {error}")
            sys.exit(102)

        config.set_step_config_overrides(args.step, args.step_config)
        # it is VERY important that the working dir be an absolute path because some
        # commands (looking at you maven) will change the context of relative paths on you
        step_runner = StepRunner(
            config=config,
            work_dir_path=os.path.abspath('step-runner-working')
        )

        try:
            if not step_runner.run_step(args.step, args.environment):
                print_error(f"Step {args.step} not successful")
                sys.exit(200)

        except Exception as error:  # pylint: disable=broad-except
            print_error(f"Fatal error calling step ({args.step}): {str(error)}")
            track = traceback.format_exc()
            print(track)
            sys.exit(300)


def init():
    """
    Notes
    -----
    See https://medium.com/opsops/how-to-test-if-name-main-1928367290cb
    """
    if __name__ == "__main__":
        sys.exit(main())


init()
